import glob
import gzip
import argparse
import wget
import os
import shutil

SUPERDARKS = {
    "superdark_FUVB_175_57650_58050_quiescent.asdf.gz": "https://stsci.box.com/shared/static/bhpjp43etyvt0mj7xye1mcj6t986qn0b.gz",
    "superdark_FUVB_175_56850_57100_active.asdf.gz": "https://stsci.box.com/shared/static/30bxwuh6k636nm14h08nu9z2a0m6162n.gz",
    "superdark_FUVB_169_56550_56800_active.asdf.gz": "https://stsci.box.com/shared/static/d5mllkh1ar7kl34o73twlccv6tzq3s5u.gz",
    "superdark_FUVB_169_57400_57700_quiescent.asdf.gz": "https://stsci.box.com/shared/static/0nl61crf5bu1g30bti3yn3hr912mjs1k.gz",
    "superdark_FUVB_163_58000_58300_quiescent.asdf.gz": "https://stsci.box.com/shared/static/xe88v32ggktwbadman10z9g40sywmyst.gz",
    "superdark_FUVB_167_57000_57300_active.asdf.gz": "https://stsci.box.com/shared/static/ual7xbt4bkmpsbsyt4kspt46fnv1yfoj.gz",
    "superdark_FUVB_167_55100_55400_quiescent.asdf.gz": "https://stsci.box.com/shared/static/b12bht6th8u5lja5v87sv6huc2jr5mlu.gz",
    "superdark_FUVB_163_56250_56420_quiescent.asdf.gz": "https://stsci.box.com/shared/static/madq09nk2rv5ugb5u2hro95nmzb0yzkp.gz",
    "superdark_FUVB_163_57050_57250_active.asdf.gz": "https://stsci.box.com/shared/static/xulxtemd8onwr8t5nb8x53va3zvf5ri5.gz",
    "superdark_FUVA_167_56520_56820_active.asdf.gz": "https://stsci.box.com/shared/static/7ifuaap3nnmm1w1jnpssxlcbasamkhnm.gz",
    "superdark_FUVA_171_57050_57300_active.asdf.gz": "https://stsci.box.com/shared/static/qug99msvg14rdqs6gqdujr276u8cov6h.gz",
    "superdark_FUVA_173_56900_57100_active.asdf.gz": "https://stsci.box.com/shared/static/5w1jxexyjdpduw9zrk4o6lh97trtb4p4.gz",
    "superdark_FUVA_167_57500_57800_quiescent.asdf.gz": "https://stsci.box.com/shared/static/0wq6duaao2ds9q5ayead3qm1oxo0gxj9.gz",
    "superdark_FUVA_163_58300_58600_active.asdf.gz": "https://stsci.box.com/shared/static/d1rmaqmu2zlzza0nc5b737d1g3icw39y.gz",
    "superdark_FUVA_169_55100_55400_quiescent.asdf.gz": "https://stsci.box.com/shared/static/n1yoeheto2ml8fgg5d98882fo8bzq6n7.gz",
    "superdark_FUVA_163_58000_58220_quiescent.asdf.gz": "https://stsci.box.com/shared/static/n8enh4s6pf2zk9w4z670js9h0l0eebhs.gz",
    "superdark_FUVA_169_55700_56010_active.asdf.gz": "https://stsci.box.com/shared/static/kosdxgsambik29cn5p618gupakpv8dv1.gz"}


def wget_superdarks(destdir):
    if not os.path.exists(destdir):
        os.makedirs(destdir)
    for item in SUPERDARKS:
        dest = os.path.join(destdir, item)
        wget.download(SUPERDARKS[item], dest)
    print(f"\n\n!!!\nIMPORTANT\n!!!\nMake sure to set the ACDC_SUPERDARKS environment variable:")
    print(f'export ACDC_SUPERDARKS="{destdir}"')
    print("For convenience, you may want to add this to your .bashrc/.cshrc/.bash_profile file")


def unzip_superdarks(destdir):
    zipped = glob.glob(os.path.join(destdir, "*gz"))
    for z_item in zipped:
        uz_item = z_item.split(".gz")[0]
        with gzip.open(z_item, "rb") as f_in, open(uz_item, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
            print("Uncompressing {} -> {}".format(z_item, uz_item))
        if os.path.isfile(uz_item):
            os.remove(z_item)


def download_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(dest="destdir",
                        help="Directory to download superdarks into")
    args = parser.parse_args()
    wget_superdarks(args.destdir)
    unzip_superdarks(args.destdir)

if __name__ == "__main__":
    download_parser()



import os
import itertools
import numpy as np
import argparse
from matplotlib import pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "niceplot.mplstyle")
plt.style.use(stylesheet)
import glob
from astropy.io import fits
from scipy import stats
import asdf

from acdc.superdark.cos_fuv_superdark import Superdark
from acdc.utils.utils import get_binning_pars, bin_coords, unbin_image

LOCAL_REFDIR = "/grp/hst/cdbs/lref"
LOCAL_DARKDIR = "/astro/sveash/cos_dark/final_superdarks"
PHA_INCLUSIVE = [2, 23]
PHA_INCL_EXCL = [2, 24]
# Colorblind-safe palette below
#COLORS = ["#1b9e77", "#d95f02", "#7570b3", "#e7298a", "#66a61e", "#a6cee3",
#          "#1f78b4", "#b2df8a", "#33a02c", "#fb9a99", "#e31a1c", "#fdbf6f",
#          "#ff7f00", "#cab2d6", "#6a3d9a", "#ffff99", "#b15928", "#a6cee3"]
COLORS = ["#9970ab", "#5aae61", "#d95f02", "#e7298a", "#1bddf2", "#1f78b4", 
          "#fb9a99", "#fdbf6f", "#ffe44b", "#b15928", "#cab2d6", "#b2df8a", 
          "#000000", "#7a7a7a", "#911cff", "#a6cee3"]

class CosImage():
    def __init__(self, filename, arr, segment, cenwave, hv, exptime, 
                 binning=None, data_binned=None):
        self.filename = filename
        self.data = arr
        self.segment = segment
        self.cenwave = cenwave
        self.hv = hv
        self.exptime = exptime
        self.binning = binning
        self.data_binned = data_binned

    
    @classmethod
    def from_flt(cls, flt_file):
        data = fits.getdata(flt_file)
        hdr0 = fits.getheader(flt_file, 0)
        hdr1 = fits.getheader(flt_file, 1)
        segment = hdr0["segment"]
        self = cls(data, flt_file, segment, hdr0["cenwave"], hdr1[f"HVLEVEL{segment[-1]}"],
                   hdr1["exptime"])
        return self


    @classmethod
    def from_asdf(cls, asdf_file):
        S = Superdark.from_asdf(asdf_file)
        if S.pha_bins != PHA_INCL_EXCL:
            S.bin_superdark(1, 1, pha_bins=PHA_INCL_EXCL)
        if S.bin_x != 1 or S.bin_y != 1:
            data_binned = S.superdarks[0]
            data = unbin_image(data_binned, S.bin_x, S.bin_y, S.xstart, S.ystart,
                               S.xend, S.yend)
        else:
            data_binned = None
            data = S.superdarks[0]
        keys = ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", "yend",
                "phastart", "phaend"]
        binning = {}
        for k in keys:
            binning[k] = getattr(S, k)
        self = cls(data, asdf_file, S.segment, None, S.hv, S.total_exptime, binning, 
                   data_binned)


    def get_background_regions(self, xtractab, cenwave):
        xtract = fits.getdata(xtractab)
        inds = np.where((xtract["segment"] == self.segment) &
                        (xtract["cenwave"] == cenwave) &
                        (xtract["aperture"] == "PSA"))
        xtract_data = xtract[inds]
        b1 = xtract_data["b_bkg1"]                                        
        b2 = xtract_data["b_bkg2"]
        height1 = xtract_data["b_hgt1"]
        height2 = xtract_data["b_hgt1"]
        m = xtract_data["slope"]
        smoothsize = xtract_data["bwidth"]
        x = np.arange(0, 16384)
        bkg1_mid = m*x + b1
        bkg1_y1 = bkg1_mid + height1/2
        bkg1_y0 = bkg1_mid - height1/2
        bkg2_mid = m*x + b2
        bkg2_y1 = bkg2_mid + height2/2
        bkg2_y0 = bkg2_mid - height2/2
        bkg1_y0 = np.round(bkg1_y0)
        bkg1_y0 = bkg1_y0.astype(int)
        bkg2_y0 = np.round(bkg2_y0)
        bkg2_y0 = bkg2_y0.astype(int)
        bkg1_y1 = np.round(bkg1_y1)
        bkg1_y1 = bkg1_y1.astype(int)
        bkg2_y1 = np.round(bkg2_y1)
        bkg2_y1 = bkg2_y1.astype(int)

#        self.bkg1_xstart = bkg1_x0
#        self.bkg1_xend = bkg1_x1
        self.bkg1_ystart = bkg1_y0
        self.bkg1_yend = bkg1_y1
#        self.bkg2_xstart = bkg2_x0
#        self.bkg2_xend = bkg2_x1
        self.bkg2_ystart = bkg2_y0
        self.bkg2_yend = bkg2_y1


def sum_bkg_region(data, bkg_lo, bkg_hi):
    assert np.shape(data)[1] == len(bkg_lo) == len(bkg_hi), "Data X-dimension must match limits of background regions"
    datasum = []
    for i in range(len(bkg_lo)):
        colsum = np.sum(data[bkg_lo[i]:bkg_hi[i], i])
        datasum.append(colsum)
    datasum = np.array(datasum)
    
    return datasum


def sum_region_ylims(data_2d, y0, y1, x0=None, x1=None):
    """
    Sum a region of a 2D image. *NOTE*: The upper limits (y1, x1) 
    are EXCLUSIVE, as is typical in python. E.g. specifying y0=9, y1=11
    will sum across rows 9 and 10 only.

    Args:
        data_2d (2D np.array): 2-dimensional array to sum over.
        y0 (int): Lower row limit.
        y1 (int): Upper row limit (EXCLUSIVE).
        x0 (int): Lower column limit.
        x1 (int): Upper column limit (EXCLUSIVE).
    """

    if x0 is None:
        x0 = 0
    if x1 is None:
        sh = np.shape(data_2d)
        x1 = sh[1] 
    datasum = np.sum(data_2d[y0:y1, x0:x1], axis=0)
    
    return datasum 


def smooth_array(data, smoothsize=100):
    bin_edges = np.arange(0, len(data), smoothsize)
    binned_y, edges, binnumber = stats.binned_statistic(np.arange(len(data)),
                                                      data,
                                                      statistic="mean",
                                                      bins=bin_edges)
    binned_x = bin_edges[:-1] + (smoothsize/2.)
    return binned_y, binned_x

def get_data(flt_custom, flt_default, superdark_custom, dark_dir=LOCAL_DARKDIR, ref_dir=LOCAL_REFDIR):
    data_custom = fits.getdata(flt_custom)
    data_default = fits.getdata(flt_default)
    segment = fits.getval(flt_custom, "segment")
    cenwave = fits.getval(flt_custom, "cenwave")
    hv = fits.getval(flt_custom, f"HVLEVEL{segment[-1]}", 1)
    xtractab_file0 = fits.getval(flt_custom, "xtractab")
    xtractab_file1 = xtractab_file0.split("$")[1]
    xtractab_file = os.path.join(ref_dir, xtractab_file1)

    xtract = fits.getdata(xtractab_file)
    inds = np.where((xtract["segment"] == segment) &
                    (xtract["cenwave"] == cenwave) & 
                    (xtract["aperture"] == "PSA"))
    xtract_data = xtract[inds]
    b1 = xtract_data["b_bkg1"]
    b2 = xtract_data["b_bkg2"]
    height1 = xtract_data["b_hgt1"]
    height2 = xtract_data["b_hgt1"]
    m = xtract_data["slope"]
    smoothsize = xtract_data["bwidth"]
    x = np.arange(0, 16384)
    bkg1_mid = m*x + b1
    bkg1_y1 = bkg1_mid + height1/2
    bkg1_y0 = bkg1_mid - height1/2
    bkg2_mid = m*x + b2
    bkg2_y1 = bkg2_mid + height2/2
    bkg2_y0 = bkg2_mid - height2/2
    bkg1_y0 = np.round(bkg1_y0)
    bkg1_y0 = bkg1_y0.astype(int)
    bkg2_y0 = np.round(bkg2_y0)
    bkg2_y0 = bkg2_y0.astype(int)
    bkg1_y1 = np.round(bkg1_y1)
    bkg1_y1 = bkg1_y1.astype(int)
    bkg2_y1 = np.round(bkg2_y1)
    bkg2_y1 = bkg2_y1.astype(int)

    superdark_lo = glob.glob(os.path.join(dark_dir, f"*{segment}*{hv}*quiescent*asdf"))
    superdark_lo = [x for x in superdark_lo if "binned" not in x]
    superdark_hi = glob.glob(os.path.join(dark_dir, f"*{segment}*{hv}*active*asdf"))
    superdark_hi = [x for x in superdark_hi if "binned" not in x]
    assert len(superdark_lo) != 0, \
        f"No quiescent superdarks found for {segment} {hv}" 
    assert len(superdark_hi) != 0, \
        f"No active superdarks found for {segment} {hv}" 
    assert len(superdark_lo) == 1, \
        f"Too many quiescent superdarks found for {segment} {hv}" 
    assert len(superdark_hi) == 1, \
        f"Too many active superdarks found for {segment} {hv}"
    superdark_lo = superdark_lo[0]
    superdark_hi = superdark_hi[0]
    S_lo = Superdark.from_asdf(superdark_lo) 
    S_lo.bin_superdark(1, 1, pha_bins=[1,31]) 
    #S_lo.bin_superdark(1, 1, pha_bins=PHA_INCL_EXCL) 
    S_hi = Superdark.from_asdf(superdark_hi) 
    S_hi.bin_superdark(1, 1, pha_bins=[1,31])
    #S_hi.bin_superdark(1, 1, pha_bins=PHA_INCL_EXCL)
    lo_exptime = S_lo.total_exptime
    hi_exptime = S_hi.total_exptime
    S_lo_flt0 = S_lo.superdarks[0] / lo_exptime
    S_hi_flt0 = S_hi.superdarks[0] / hi_exptime
    assert np.shape(S_lo_flt0) == np.shape(S_hi_flt0), \
        "Quiescent and active superdark dimensions do not match"
    xstart = S_lo.bin_xstart
    xend = S_lo.bin_xend
    ystart = S_lo.bin_ystart
    yend = S_lo.bin_yend
    S_lo_flt = np.zeros(16777216).reshape(1024, 16384)
    S_hi_flt = np.zeros(16777216).reshape(1024, 16384)
    S_lo_flt[ystart:yend, xstart:xend] = S_lo_flt0
    S_hi_flt[ystart:yend, xstart:xend] = S_hi_flt0

    bkg1_sum_lo = sum_bkg_region(S_lo_flt, bkg1_y0, bkg1_y1)
    bkg1_lo,_ = smooth_array(bkg1_sum_lo)
    bkg2_sum_lo = sum_bkg_region(S_lo_flt, bkg2_y0, bkg2_y1)
    bkg2_lo,_ = smooth_array(bkg2_sum_lo)
    bkg1_sum_hi = sum_bkg_region(S_hi_flt, bkg1_y0, bkg1_y1)
    bkg1_hi,_ = smooth_array(bkg1_sum_hi)
    bkg2_sum_hi = sum_bkg_region(S_hi_flt, bkg2_y0, bkg2_y1)
    bkg2_hi,_ = smooth_array(bkg2_sum_hi)

    bkg1_sum_custom = sum_bkg_region(data_custom, bkg1_y0, bkg1_y1)
    bkg1_custom,_ = smooth_array(bkg1_sum_custom)
    bkg2_sum_custom = sum_bkg_region(data_custom, bkg2_y0, bkg2_y1)
    bkg2_custom,_ = smooth_array(bkg2_sum_custom)
    bkg1_sum_default = sum_bkg_region(data_default, bkg1_y0, bkg1_y1)
    bkg1_default,_ = smooth_array(bkg1_sum_default)
    bkg2_sum_default = sum_bkg_region(data_default, bkg2_y0, bkg2_y1)
    bkg2_default, binned_x = smooth_array(bkg2_sum_default)
    
    af = asdf.open(superdark_custom)
    pha_str = f"pha{PHA_INCL_EXCL[0]}-{PHA_INCL_EXCL[1]}"
    dark_im = af.tree[pha_str]
    dark_im_flt0 = dark_im / af.tree["total_exptime"]
    binning = get_binning_pars(af)
    dark_im_flt_perpixel = dark_im_flt0 / binning["bin_y"] / binning["bin_x"]
    dark_im_flt = np.zeros(16777216).reshape(1024, 16384)
    xs = np.arange(binning["xstart"], binning["xend"], binning["bin_x"])
    ys = np.arange(binning["ystart"], binning["yend"], binning["bin_y"])
    for i in range(len(xs)-1): 
        for j in range(len(ys)-1):
            dark_im_flt[ys[j]:ys[j+1], xs[i]:xs[i+1]] = dark_im_flt_perpixel[j,i]
    dark_x = np.arange(16384)
    
    bkg1_dark = sum_bkg_region(dark_im_flt, bkg1_y0, bkg1_y1)
    bkg2_dark = sum_bkg_region(dark_im_flt, bkg2_y0, bkg2_y1)
    
    return binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, \
        bkg1_lo, bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x
    #return np.arange(16384), bkg1_sum_default, bkg2_sum_default, bkg1_sum_custom, bkg2_sum_custom, bkg1_sum_lo, bkg2_sum_lo, bkg1_sum_hi, bkg2_sum_hi


def plot_data(x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, bkg1_lo, 
              bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x, targname,
              flt_default):
    fig, axes0 = plt.subplots(3, 1, figsize=(9, 12))
    fig.subplots_adjust(hspace=0.3)
    segment = fits.getval(flt_default, "segment")
    rootname = fits.getval(flt_default, "rootname").lower()
    hv = fits.getval(flt_default, f"HVLEVEL{segment[-1]}", 1)
    date = fits.getval(flt_default, "date-obs", 1)
    axes = axes0.flatten()
    
    bkg1 = {"default": bkg1_default, "custom": bkg1_custom, "lo": bkg1_lo, "hi": bkg1_hi, "superdark": bkg1_dark}
    bkg2 = {"default": bkg2_default, "custom": bkg2_custom, "lo": bkg2_lo, "hi": bkg2_hi, "superdark": bkg2_dark}
    av = {"default": (bkg1_default + bkg2_default) / 2.,
          "custom": (bkg1_custom + bkg2_custom) / 2.,
          "lo": (bkg1_lo + bkg2_lo) / 2.,
          "hi": (bkg1_hi + bkg2_hi) / 2.,
          "superdark": (bkg1_dark + bkg2_dark) / 2.}
    dicts = [bkg1, bkg2, av]

    titles = ["Background Region 1", "Background Region 2", "Average"]
    alpha = 0.7
    for i in range(3):
        ax = axes[i]
        ax.hlines(y=0, xmin=-1000, xmax=17384, colors="lightgrey", linestyles="--")
        d = dicts[i]
        ax.set_ylabel("Counts/s")
        ax.plot(x, d["default"],        COLORS[0], alpha=alpha, lw=1.7, label="Default FLT")
        ax.plot(x, d["custom"],         COLORS[1], alpha=alpha, lw=1.7, label="Custom FLT", zorder=100)
        ax.plot(x, d["lo"],             COLORS[2], alpha=alpha, lw=1.7, label="Lo Dark")
        ax.plot(x, d["hi"],             COLORS[3], alpha=alpha, lw=1.7, label="Hi Dark")
        ax.plot(dark_x, d["superdark"], COLORS[4], alpha=alpha, lw=1.7, label="Custom Dark")
        ax.set_ylim(-.0001, 0.0004)
        ax.set_xlim(-500, 16884)
        ax.set_title(titles[i])
    axes[0].legend()
    figname = f"{rootname}_{segment}_flt_comp.png"
    title = f"{segment}; HV={hv}; {date} ({rootname})"
    fig.suptitle(title, size=20, y=.94)
    fig.savefig(figname)
    print(f"Wrote {figname}")


def compare_backgrounds(flt_default, flt_custom, targname, superdark_custom):
    binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, bkg1_lo, \
        bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x = get_data(flt_custom, flt_default, superdark_custom)
    plot_data(binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, \
        bkg1_lo, bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x, targname, flt_default)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--default",
                        help="Name of default CalCOS FLT file")
    parser.add_argument("-c", "--custom",
                        help="Name of custom-corrected FLT file")
    parser.add_argument("-t", "--targname",
                        help="Name of target")
    parser.add_argument("-s", "--superdark_custom",
                        help="Name of combined superdark from ACDC")
    args = parser.parse_args()
    compare_backgrounds(args.default, args.custom, args.targname, args.superdark_custom)





from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column
from sqlalchemy import Float, ForeignKey, Integer, String
from sqlalchemy.types import DECIMAL as Decimal

Base = declarative_base()

class DarkEvents(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkevents"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv163(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv163"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv167(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv167"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv169(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv169"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv171(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv171"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv173(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv173"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv175(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv175"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())

class DarkEventsHv178(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshv178"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())


class DarkEventsHvOther(Base):
    """Object-relational mapping for the DarkEvents table."""

    __tablename__ = "darkeventshvother"
    id = Column(Integer, primary_key=True, nullable=False)
    xcorr = Column(Decimal(12,5))
    ycorr = Column(Decimal(12,5))
    pha = Column(Integer())
    mjd = Column(Decimal(precision='20,11'))
    hv = Column(Integer())
    segment = Column(String(5))
    filename = Column(String(30))
    proposid = Column(Integer())



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column
from sqlalchemy import Float, ForeignKey, Integer, String

Base = declarative_base()

class Darks(Base):
    """Object-relational mapping for the Darks table."""

    __tablename__ = "darks"
    id = Column(Integer, primary_key=True, nullable=False)
    filename = Column(String(30))
    segment = Column(String(5))
    expstart = Column(Float())
    exptime = Column(Float())
    hv = Column(Integer())
    latitude = Column(Float())
    longitude = Column(Float())
    saa_distance= Column(Float())
    region = Column(String(15))
    region_area = Column(Integer())
    xcorr_min = Column(Integer())
    xcorr_max = Column(Integer())
    ycorr_min = Column(Integer())
    ycorr_max = Column(Integer())
    solar_flux = Column(Float())
    fileloc = Column(String(200))
    dark_pha0 = Column(Integer())
    dark_pha1 = Column(Integer())
    dark_pha2 = Column(Integer())
    dark_pha3 = Column(Integer())
    dark_pha4 = Column(Integer())
    dark_pha5 = Column(Integer())
    dark_pha6 = Column(Integer())
    dark_pha7 = Column(Integer())
    dark_pha8 = Column(Integer())
    dark_pha9 = Column(Integer())
    dark_pha10 = Column(Integer())
    dark_pha11 = Column(Integer())
    dark_pha12 = Column(Integer())
    dark_pha13 = Column(Integer())
    dark_pha14 = Column(Integer())
    dark_pha15 = Column(Integer())
    dark_pha16 = Column(Integer())
    dark_pha17 = Column(Integer())
    dark_pha18 = Column(Integer())
    dark_pha19 = Column(Integer())
    dark_pha20 = Column(Integer())
    dark_pha21 = Column(Integer())
    dark_pha22 = Column(Integer())
    dark_pha23 = Column(Integer())
    dark_pha24 = Column(Integer())
    dark_pha25 = Column(Integer())
    dark_pha26 = Column(Integer())
    dark_pha27 = Column(Integer())
    dark_pha28 = Column(Integer())
    dark_pha29 = Column(Integer())
    dark_pha30 = Column(Integer())
    dark_pha31 = Column(Integer())

class Solar(Base):
    """Object-relational mapping for the Solar table."""

    __tablename__ = "solar"
    id = Column(Integer, primary_key=True, nullable=False)
    time = Column(Float())
    flux = Column(Float())




import numpy as np
import argparse
from astropy.io import fits
import os
import glob

from acdc.database.query_cos_dark import files_by_mjd

def sum_data(mjdstart, mjdend):
    df = files_by_mjd(mjdstart, mjdend)

    total_time_a = 0
    total_time_b = 0
    sumdark_a = np.zeros(16384*1024).reshape(1024, 16384)
    sumdark_b = np.zeros(16384*1024).reshape(1024, 16384)
    for i in range(len(df)):
        fileloc = df.iloc[i]["fileloc"]
        segment = df.iloc[i]["segment"]
        filename = os.path.basename(fileloc)
        pid = os.path.dirname(fileloc).split("/")[-1]
        cosmofile = glob.glob(os.path.join("/grp/hst/cos2/cosmo/", pid, filename.replace("corrtag", "counts")+"*"))[0]
        print(cosmofile)
        exptime = fits.getval(cosmofile, "exptime", 1)
        data = fits.getdata(cosmofile)
        if segment == "FUVA":
            total_time_a += exptime
            sumdark_a += data
        else:
            total_time_b += exptime
            sumdark_b += data

    superdark_a = sumdark_a / total_time_a
    superdark_b = sumdark_b / total_time_b
    return superdark_a, superdark_b, total_time_a, total_time_b

def make_superdark(mjdstart, mjdend):
    superdark_a, superdark_b, total_time_a, total_time_b = sum_data(mjdstart, mjdend)
    
    hdr = fits.Header()
    hdr["MJDSTART"] = mjdstart
    hdr["MJDEND"] = mjdend
    primary = fits.PrimaryHDU(header=hdr)
    
    hdr1 = fits.Header()
    hdr1["EXTNAME"] = "SUPERDARK_FUVA"
    hdr1["EXPTIME"] = total_time_a
    sci1 = fits.ImageHDU(superdark_a, header=hdr1, name="SUPERDARK_FUVA")
    
    hdr2 = fits.Header()
    hdr2["EXTNAME"] = "SUPERDARK_FUVA"
    hdr2["EXPTIME"] = total_time_b
    sci2 = fits.ImageHDU(superdark_b, header=hdr2, name="SUPERDARK_FUVB")

    hdu = fits.HDUList([primary, sci1, sci2])
    outname = f"superdark_{mjdstart}_{mjdend}.fits"
    hdu.writeto(outname, overwrite=True)
    print(f"Wrote file {outname}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(dest="mjdstart", 
                        help="Starting MJD time to select files")
    parser.add_argument(dest="mjdend", 
                        help="Ending MJD time to select files")
    args = parser.parse_args()

    make_superdark(args.mjdstart, args.mjdend)



import os
import glob
import argparse
import datetime

from ullyses.generic_coadd_wrapper import coadd_files

def coadd_cos_spectra(infiles, outdir=None, outfile=None, clobber=False):
    """Coadd multiple 1-D COS spectra of the same grating.
    Args:
        grating (str): COS grating name.
        indir (str): Path that holds all data to be coadded. All data in this
            directory that match grating will be coadded.
    Returns:
        None
    """

    if outdir is None:
        now = datetime.datetime.now()
        outdir = f"{now.strftime('%d%b%Y_%H%M%S')}_coadd"  
    coadd_files(infiles, outdir, outfile, clobber)


def find_files(indir):
    allfiles = glob.glob(os.path.join(indir, "*x1d.fits"))
    return allfiles

def coadd_parser():
    """
    Copied from ullyses.ullyses_coadd_abut_wrapper
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--indir",
                        default="./",
                        help="Directory(ies) with data to combine")
    parser.add_argument("-o", "--outdir", default=None,
                        help="Directory for output HLSPs")
    parser.add_argument("--outfile", default=None,
                        help="Name of output coadded file")
    parser.add_argument("-c", "--clobber", default=False,
                        action="store_true",
                        help="If True, overwrite existing products")
    args = parser.parse_args()

    infiles = find_files(args.indir)
    coadd_cos_spectra(infiles, args.outdir, args.outfile, args.clobber)


if __name__ == "__main__":
    coadd_parser()



import yaml
from sqlalchemy.orm import load_only
import pandas as pd

from .connect_db import load_connection
from .schema import Solar, Darks

TESTING = False

def all_darks(dbname="cos_dark"):
    """
    Query the Darks table to get darks for all time for all PHAs.

    Args:
        dbname (str): The location of the SQLite database, with full path, e.g.
            /path/to/cos_dark.db 
            If in the current directory, do not include . or ./ 
    Returns:
        df (:obj:`pandas.DataFrame`): Query results.
    """
    
	# Connect to database
    session, engine = load_connection(dbname)

    # Define columns to return from database
    cols = ["expstart", "solar_flux", "latitude", "longitude", "segment", "hv", "region", "saa_distance"]
    cols += [f"dark_pha{x}" for x in range(0,32)]
    if TESTING is True: 
        cols = ["expstart", "latitude", "longitude", "solar_flux", "dark_pha11"]

    cols_attr = [getattr(Darks, f) for f in cols]
    # Execute SELECt query with WHERE statements
    query = session.query(Darks).options(load_only(*cols_attr))
#                .filter(Darks.region == "inner")\
#                .filter(Darks.segment == "FUVA")
    if TESTING is True:
        query = query.filter(Darks.filename == "le0a3mcuq_corrtag_a.fits")
    results = query.all()

    # Convert query results (list of attributes) to pandas dataframe
    d = {col: [getattr(x, col) for x in results] for col in cols}
    df = pd.DataFrame(d)
    
    return df

def files_by_mjd(mjdstart, mjdend, segment="FUVA", hv=167, morecols=[],
                 dbname="cos_dark"):
    
	# Connect to database
    session, engine = load_connection(dbname)

    cols = ["fileloc"]

    if hv == "*" or hv == None:
        query = session.query(Darks.fileloc.distinct())\
                .filter(Darks.expstart >= mjdstart)\
                .filter(Darks.expstart < mjdend)\
                .filter(Darks.segment == segment)\
                .order_by(Darks.expstart)
    else:
        query = session.query(Darks.fileloc.distinct())\
                .filter(Darks.expstart >= mjdstart)\
                .filter(Darks.expstart < mjdend)\
                .filter(Darks.segment == segment)\
                .filter(Darks.hv == hv)\
                .order_by(Darks.expstart)
    results = query.all()
    
    d = {cols[i]: [x[i] for x in results] for i in range(len(cols))}
    df = pd.DataFrame(d)

    return df

def counts_by_mjd(mjdstart, mjdend, morecols=[],
                  dbname="cos_dark"):
    
	# Connect to database
    session, engine = load_connection(dbname)
    cols = ["expstart", "solar_flux", "latitude", "longitude", "segment", "hv", "region", "saa_distance"]
    cols += [f"dark_pha{x}" for x in range(0,32)]
    cols += morecols
    cols_attr = [getattr(Darks, f) for f in cols]

    query = session.query(Darks).options(load_only(*cols_attr))\
                .filter(Darks.expstart >= mjdstart)\
                .filter(Darks.expstart < mjdend)
    results = query.all()

    d = {col: [getattr(x, col) for x in results] for col in cols}
    df = pd.DataFrame(d)

    return df




import numpy as np
import argparse

from acdc.superdark.cos_fuv_superdark import Superdark

def all_hvs():
    S = Superdark(hv=163, segment= "FUVA", mjdstarts=[58000], mjdends=[58220], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_163_58000_58220_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=163, segment= "FUVA", mjdstarts=[58300], mjdends=[58600], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_163_58300_58600_active.asdf")
    S.create_superdark()
    S = Superdark(hv=167, segment= "FUVA", mjdstarts=[57500], mjdends=[57800], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_167_57500_57800_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=167, segment= "FUVA", mjdstarts=[56520], mjdends=[56820], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_167_56520_56820_active.asdf")
    S.create_superdark()
    S = Superdark(hv=169, segment= "FUVA", mjdstarts=[55100], mjdends=[55400], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_169_55100_55400_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=169, segment= "FUVA", mjdstarts=[55700], mjdends=[56010], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_169_55700_56010_active.asdf")
    S.create_superdark()
    S = Superdark(hv=171, segment= "FUVA", mjdstarts=[57050], mjdends=[57300], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_171_57050_57300_active.asdf")
    S.create_superdark()
    S = Superdark(hv=173, segment= "FUVA", mjdstarts=[56900], mjdends=[57100], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVA_173_56900_57100_active.asdf")
    S.create_superdark()
    S = Superdark(hv=163, segment= "FUVB", mjdstarts=[56250], mjdends=[56420], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_163_56250_56420_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=163, segment= "FUVB", mjdstarts=[58000], mjdends=[58300], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_163_58000_58300_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=163, segment= "FUVB", mjdstarts=[57050], mjdends=[57250], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_163_57050_57250_active.asdf")
    S.create_superdark()
    S = Superdark(hv=167, segment= "FUVB", mjdstarts=[55100], mjdends=[55400], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_167_55100_55400_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=167, segment= "FUVB", mjdstarts=[57000], mjdends=[57300], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_167_57000_57300_active.asdf")
    S.create_superdark()
    S = Superdark(hv=169, segment= "FUVB", mjdstarts=[57400], mjdends=[57700], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_169_57400_57700_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=169, segment= "FUVB", mjdstarts=[56550], mjdends=[56800], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_169_56550_56800_active.asdf")
    S.create_superdark()
    S = Superdark(hv=175, segment= "FUVB", mjdstarts=[57650], mjdends=[58050], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_175_57650_58050_quiescent.asdf")
    S.create_superdark()
    S = Superdark(hv=175, segment= "FUVB", mjdstarts=[56850], mjdends=[57100], bin_pha=1, phastart=1, phaend=31, 
                  outfile="superdark_FUVB_175_56850_57100_active.asdf")
    S.create_superdark()

if __name__ == "__main__":
#    parser = argparse.ArgumentParser()
#    parser.add_argument("--hv", type=int, help="HV setting of interest")
#    parser.add_argument("--segment", help="Segment of interest")
#    parser.add_argument("--mjdstarts", nargs="+", help="MJD start dates")
#    parser.add_argument("--mjdends", nargs="+", help="MJD end dates") 
#    parser.add_argument("--phabin", default=1,
#                        help="Size of PHA binning")
#    parser.add_argument("--phastart", default=1, type=int,
#                        help="Minimum PHA to consider")
#    parser.add_argument("--phaend", default=31, type=int,
#                        help="Maximum PHA to consider")
#    parser.add_argument("--pha_bins", nargs="*", default=None,
#                        help="Explicit PHA bins") 
#    parser.add_argument("-o", "--outfile", default=None,
#                        help="Name of output superdark")
#    args = parser.parse_args()
#
#    mjdstarts = [int(x) for x in args.mjdstarts]
#    mjdends = [int(x) for x in args.mjdends]
#    if args.pha_bins is not None:
#        pha_bins = np.array([int(x) for x in args.pha_bins])
#    S = Superdark(hv=args.hv, segment=args.segment, mjdstarts=mjdstarts, 
#                  mjdends=mjdends, bin_pha=args.phabin, phastart=args.phastart, 
#                  phaend=args.phaend, pha_bins=pha_bins, outfile=args.outfile)
#    S.create_superdark()
    all_hvs()



import os
import yaml
import getpass
from sqlalchemy.pool import NullPool
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

def load_connection(dbname, echo=False):
    """
    Return session, base, and engine objects for connecting to the database.

    Args:
        dbsettings (dict): Dictionary listing database type and all required
            connection parameters.
            Params will get plugged into the full connection string of form:
            `dialect+driver://username:password@host:port/database`
        echo (Bool): If True, log all statements to STDOUT. Defaults to False.

    Returns:
        session (:obj:`Session()` instance): Provides a holding zone 
            for all objects loaded or associated with the database.
        engine (:obj:`engine` object): Provides a source of database 
            connectivity and behavior.
    """

    cwd = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(cwd, "settings.yaml"), "r") as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)
        dbsettings = settings["dbsettings"][dbname]

    if dbsettings["dbtype"] == "sqlite":
        connection_string = f"sqlite:///{dbsettings['loc']}"
    elif dbsettings["dbtype"] == "mysql":
        pswd = getpass.getpass(f"Enter password for {dbsettings['dbname']}: ")
        connection_string = f"mysql+pymysql://{dbsettings['user']}:{pswd}@{dbsettings['host']}:{dbsettings['port']}/{dbsettings['dbname']}"
    engine = create_engine(connection_string, echo=echo, poolclass=NullPool)
    Session = sessionmaker(bind=engine)
    session = Session()

    return session, engine




import os
import datetime
import argparse
from matplotlib import pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "niceplot.mplstyle")
plt.style.use(stylesheet)
import numpy as np
from astropy.io import fits
import glob
from scipy import stats
import pandas as pd

# Eventually put Svea's plots here

def compare_x1ds(custom_coadd_txt, default_coadd_txt, targ=None, outfile=None, 
                 save=True):

    custom = pd.read_csv(custom_coadd_txt, names=["i", "wl", "flux", "err"], 
                            header=None, delim_whitespace=True)   
    default = pd.read_csv(default_coadd_txt, names=["i", "wl", "flux", "err"], 
                            header=None, delim_whitespace=True)
    
    fig, axes0 = plt.subplots(3, 1, figsize=(20, 16), sharex=True)
    axes = axes0.flatten()
    
    axes[0].plot(custom["wl"], custom["flux"], "crimson", alpha=0.8, label="Custom")
    axes[0].plot(default["wl"], default["flux"], "royalblue", alpha=0.8, label="Default")
    axes[0].set_xlabel("Wavelength [A]")
    axes[0].set_ylabel("Flux [ergs/s/cm^2/A]")
    axes[0].legend(loc="lower right")

    axes[1].plot(custom["wl"], custom["err"], "crimson", alpha=0.8, label="Custom")
    axes[1].plot(default["wl"], default["err"], "royalblue", alpha=0.8, label="Default")
    axes[1].set_xlabel("Wavelength [A]")
    axes[1].set_ylabel("Error [ergs/s/cm^2/A]")
    axes[1].legend(loc="lower right")

    axes[2].plot(custom["wl"], custom["flux"]/custom["err"], "crimson", 
                 alpha=0.8, label="Custom")
    axes[2].plot(default["wl"], default["flux"]/default["err"], "royalblue", 
                 alpha=0.8, label="Default")
    axes[2].set_xlabel("Wavelength [A]")
    axes[2].set_ylabel("S/N")
    axes[2].legend(loc="lower right")

    fig.suptitle(targ)
    plt.tight_layout()

    if save is True:
        if outfile is None:
            if targ is not None:
                outfile = f"{targ}_x1d_comp.png"
            else:
                now = datetime.datetime.now()
                outfile = f"{now.strftime('%d%b%Y_%H%M%S')}_x1d_comp.png"
        plt.savefig(outfile, )
        print(f"Saved {outfile}")
    else:
        print("\nClose figure when finished")
        plt.show()

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--custom",
                        help="Coadd txt file for custom dark-corrected data")
    parser.add_argument("-d", "--default",
                        help="Coadd txt file for default dark-corrected data")
    parser.add_argument("-t", "--targ", default=None,
                        help="Name of target being compared")
    parser.add_argument("-o", "--outfile", default=None,
                        help="Name of output figure")
    parser.add_argument("--interact", default=False, action="store_true",
                        help="If True, open plot interactive window and do not save figure")
    args = parser.parse_args()
    save = not args.interact

    compare_x1ds(args.custom, args.default, args.targ, args.outfile, save)



"""
Given COS science datasets, compare the background regions in the science
exposures to a low/quiescent and high/active superdark.

Command line-arguments:
    -d or --datatdir
        The path to science corrtags
    --hv 
        If datadir contains corrtags of multiple HVs, you can use --hv to use
        just one specific HV value.
    --segment
        If datadir contains corrtags of multiple segments, you can use --segment
        to use just one specific segment.
    --lo
        The name of the low activity superdark.
    --hi
        The name of the high activity superdark.
    -o or --outdir
        The name of the output directory where products will be written.
    --binned
        Using this argument indicates that the supplied superdarks are already
        binned and should not be binned any further.
"""
import datetime
from collections import defaultdict
import itertools
import copy
import argparse
import os
import glob
import asdf
from scipy.optimize import minimize
from astropy.io import fits
import numpy as np
import matplotlib.pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "analysis", "niceplot.mplstyle")
plt.style.use(stylesheet)
from matplotlib.backends.backend_pdf import PdfPages
import warnings
from astropy.units import UnitsWarning
warnings.filterwarnings("ignore", category=UnitsWarning)
import pandas as pd
pd.options.mode.chained_assignment = None

from acdc.superdark.cos_fuv_superdark import Superdark, get_gsag_holes
from acdc.database.calculate_dark import get_aperture_region
from acdc.analysis.compare_backgrounds import smooth_array
from acdc.utils.utils import bin_coords, timefunc, get_psa_wca


# Hardcoded constant values
RESEL = [6, 10] # FUV resolution element size
# These are the PHAs that CalCOS uses for calibration, where the 
# lower and upper bounds are both inclusive 
PHA_INCLUSIVE = [2, 23] 
# These are the PHAs that CalCOS uses for calibration, where the 
# lower bound is inclusive and upper bounds is exclusive. This is
# the logic python uses!!
PHA_INCL_EXCL = [2, 24]
# Colorblind-safe palette below
COLORS = ["#9970ab", "#5aae61", "#d95f02", "#e7298a", "#66a61e", "#a6cee3", 
          "#1f78b4", "#b2df8a", "#33a02c", "#fb9a99", "#e31a1c", "#fdbf6f", 
          "#ff7f00", "#cab2d6", "#6a3d9a", "#ffff99", "#b15928", "#a6cee3"]
# Hydrogen Lyman alpha wavelength limits
LYA_LO = 1213.67 # Angstroms
LYA_HI = 1217.67 # Angstroms
# Oxygen I wavelength limits
OI_LO = 1300.5
OI_HI = 1307.0

def measure_gsag(corrtag, row_threshold=.3):
    if fits.getval(corrtag, "life_adj") != 1:
        lp1_interpolate = False
    else:
        lp1_interpolate = None
    do_interpolate = True
    gsagtab = fits.getval(corrtag, "gsagtab")
    if "$" in gsagtab:
        lref = os.environ["lref"]
        gsagtab = os.path.join(lref, gsagtab.split("$")[1])
    segment = fits.getval(corrtag, "segment")
    hv = fits.getval(corrtag, f"HVLEVEL{segment[-1]}", 1)
    mjdend = fits.getval(corrtag, "EXPEND", 1)
    gsag_holes = get_gsag_holes(gsagtab, segment, hv)
    gsag_df = gsag_holes[gsag_holes["DATE"] < mjdend]
    if len(gsag_df) == 0:
        return do_interpolate, lp1_interpolate
    def prod(x0, y0, dx, dy):
        out = list(itertools.product(range(x0, x0+dx+1), range(y0, y0+dy+1)))
        return out
    coords = []
    for i in range(len(gsag_df)):
        s = gsag_df.iloc[i]
        c = prod(s["X0"], s["Y0"], s["DX"], s["DY"])
        coords += c

    #gsag_df["COORDS"] = gsag_df.apply(lambda row: prod(row['X0'], row['Y0'], row['DX'], row['DY']), axis=1)
    #coords0 = gsag_df['COORDS'].to_numpy()
    #coords = []
    #for item in coords0: #each element is a list
    #    coords += item
#    x = [x[0] for x in coords]
#    y = [x[1] for x in coords]
#    plt.plot(x, y, "ko")
#    plt.show()
    coords = list(set(coords))
    coords_x = np.array([x[0] for x in coords])
    coords_y = np.array([x[1] for x in coords])
    # Active Area limits from BRFTAB x1u1459il_brf.fits (essentially static)
    aa_lims = {"FUVA": (1060, 296, 15250, 734), #xstart, ystart, xend, yend
               "FUVB": (809,  360, 15182, 785)} #xstart, ystart, xend, yend
    seg_aa_lims = aa_lims[segment]
    pix_row = seg_aa_lims[2] - seg_aa_lims[0]
    for y in range(seg_aa_lims[1], seg_aa_lims[3]+1):
        inds = np.where((coords_y == y) & (coords_x > seg_aa_lims[0]) & (coords_x < seg_aa_lims[2]))
        frac = len(inds[0])/pix_row
        if frac >= row_threshold:
            do_interpolate = False
            break

    print(f"Interpolate={do_interpolate} for {corrtag}")
    return do_interpolate, lp1_interpolate

#    ngsag = sum(gsag_df["DX"] * gsag_df["DY"])
#    npix = 16777216 #16384*1024
#    det_threshold = (row_threshold*16384)/npix
#    gsag_frac = ngsag/npix
#    do_interpolate = True
#    if gsag_frac >= det_threshold:
#        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
#        print(corrtag)
#        print(f"Number of gain-sagged pixels exceeds threshold!")
#        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
#        do_interpolate = False

#    data = fits.getdata(corrtag)
#    gsaginds = np.where(data["dq"]&8192 == 8192)
#    x = data["xfull"][gsaginds]
#    y = data["yfull"][gsaginds]
#    x = x.astype(int)
#    y = y.astype(int)
#    coords = [(x[i], y[i]) for i in range(len(x))]
#    uniq_coords = list(set(coords))
#    ngsag = len(uniq_coords)
#    nevents = len(data["xfull"])
#    print((ngsag/nevents)*100.)
#    npix = 16777216 #16384*1024
#    det_threshold = (row_threshold*16384)/npix
#    if ngsag/npix >= det_threshold:
#        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
#        print(f"Number of gain-sagged pixels exceeds threshold!")
#        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")

#@timefunc
def fun_opt(coeffs, darks, binned_sci, excluded_rows):
    """TODO- fill in

    Arguments:
        coeffs (list): The output coefficients.
        darks (list): A list of all superdarks to model, typicall [quiescent_dark, active_dark].
        binned_sci (array): Binned science image.
        excluded_rows (array): List of rows to exclude from modeling. Correspond
            to the PSA and WCA regions.
    Returns:
        cval (list): Coefficients? TODO
    """

    combined_dark = linear_combination(darks, coeffs)
    cval = c_stat(combined_dark, binned_sci, excluded_rows)
    # Uncomment the lines below for debugging on the convergence process
#    print(coeffs)
#    print(cval)
#    print("")
    
    return cval


#@timefunc
def linear_combination(darks, coeffs):
    """Scale and combine superdarks.

    Scale multiple superdarks (typically 1 quiescent and 1 active superdark),
    then add them together to create a scaled model superdark for a particular
    science exposure.

    Arguments:
        darks (list): List of superdark images to scale and combine.
        coeffs (list): List of scale factors to apply to each superdark.
    Returns:
        combined (array): Scaled and combined model superdark.
    """

    combined = darks[0] * coeffs[0]
    for i in range(1, len(darks)):
        combined += darks[i] * coeffs[i]
    return combined 


def check_superdarks(binned_superdarks):
    """Ensure all input superdarks were binned with identical bin sizes.

    Arguments: 
        binned_superdarks (list or array-like): List of binned superdarks.
    """

    bad = False
    keys = ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", "yend",
            "phastart", "phaend"]
    dark_vals = defaultdict(list)
    for darkfile in binned_superdarks:
        dark_af = asdf.open(darkfile)
        for k in keys:
            dark_vals[k].append(dark_af[k])
        dark_af.close()
    for k in keys:
        if len(set(dark_vals[k])) != 1:
            print(f"WARNING!!!! Key {k} does not match for all superdarks")
            bad = True
    if bad is True:
        for darkfile in binned_superdarks:
            print(darkfile)
        raise KeyError("Not all input superdarks are binned identically")


def get_binning_pars(af):
    """Return spatial and PHA binning information for a superdark. 
    
    Arguments:
        af (AsdfFile): ASDF file to extract binning information from.
    Returns:
        binning (dict): Dictionary that describes the binning information
            in both the spatial and PHA dimensions.
    """
    
    keys = ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", "yend",
            "phastart", "phaend"]
    binning = {}
    for k in keys:
        binning[k] = af[k]

    return binning

def bin_science(corrtag, b, fact=1, exclude_airglow=False):
    """
    Given a corrtag with lists of events as a function of X, Y, and PHA,
    bin the data into an image using the same bin sizes as the superdark.

    Arguments:
        corrtag (str): Corrtag to bin.
        b (dict): Dictionary that describes the binning information
            in both the spatial and PHA dimensions. (Only spatial is used)
    Returns:
        binned (array): Binned science image.
        nevents (array): Number of events in each binned pixel.
    """
    
    aperture = "PSA"
    segment = fits.getval(corrtag, "segment")
    cenwave = fits.getval(corrtag, "cenwave")
    lp = fits.getval(corrtag, "life_adj")
    aperture_regions = get_aperture_region(cenwave=cenwave, aperture=aperture, segments=[segment], life_adj=[lp])
    box = aperture_regions[segment][f"lp{lp}_{aperture.lower()}_{cenwave}"]
    xmin0, xmax0, ymin0, ymax0 = box
    data = fits.getdata(corrtag)
    inds0 = np.where(
                     (data["pha"] >= b["phastart"]) & 
                     (data["pha"] < b["phaend"]) &
                     (data["xcorr"] >= b["xstart"]) &
                     (data["xcorr"] < b["xend"]) & 
                     (data["ycorr"] >= b["ystart"]) &
                     (data["ycorr"] < b["yend"]))
    filtered0 = data[inds0]
    if exclude_airglow is True:
        print(f"Excluding pixels affected by LyAlpha & OI airglow") 
        bad = (filtered0["wavelength"] > LYA_LO) &\
              (filtered0["wavelength"] < LYA_HI) &\
              (filtered0["wavelength"] > OI_LO) &\
              (filtered0["wavelength"] < OI_HI)
        #      (filtered0["ycorr"] > ymin0) &\
        #      (filtered0["ycorr"] < ymax0)
        inds1 = ~bad
        filtered = filtered0[inds1]
    else:
        filtered = filtered0

    ydim = int((b["yend"] - b["ystart"]) / b["bin_y"])
    xdim = int((b["xend"] - b["xstart"]) / b["bin_x"])
    binned = np.zeros((ydim, xdim*fact))
    nevents = np.zeros((ydim, xdim*fact))
    
    xbinned, ybinned = bin_coords(filtered["xcorr"], filtered["ycorr"],
                            b["bin_x"], b["bin_y"], b["xstart"], b["ystart"],
                            make_int=True)
    
    for i in range(len(xbinned)):
        x = xbinned[i]
        y = ybinned[i]
        binned[y,x] += filtered["epsilon"][i]
        nevents[y,x] += 1

    return binned, nevents


#@timefunc
def c_stat(combined_dark, binned_sci, excluded_rows):
    """TODO document

    Arguments:
        combined_ark (array): Scaled and combined model superdark.
        binned_sci (array): Binned science image.
        excluded_rows (array): List of rows to exclude. Corresponds
            to the PSA and WCA regions.
    
    Returns:
        csum (float): TODO
    """

    csum = 0.
    for y in range(combined_dark.shape[0]):
        for x in range(combined_dark.shape[1]):
            if binned_sci[y,x] > 0 and y not in excluded_rows and combined_dark[y,x] > 0:
                csum = csum + 2. * (combined_dark[y,x] - binned_sci[y,x] + binned_sci[y,x] * (np.log(binned_sci[y,x]) - np.log(combined_dark[y,x])))
            elif y not in excluded_rows and combined_dark[y,x] > 0:
                csum += np.abs(combined_dark[y,x]) * 2.
            elif y not in excluded_rows and combined_dark[y,x] <= 0:
                csum += np.inf

    return csum


def check_existing(filename, overwrite=False):
    if os.path.exists(filename) and overwrite is False:
        print(f"WARNING: File {filename} already exists and overwrite is False, skipping...")
        return True
    else:
        return False


def predict_dark(corrtags, superdarks, segment=None, hv=None, 
                 outdir=".", binned=False, overwrite=False, exclude_airglow=False,
                 x_bin=RESEL[0]*2, y_bin=RESEL[1]*2):
    """Model the superdark for each input science corrtag.

    Use the active and quiescent superdarks of the
    appropriate segment+HV combination to determine the best predicted
    superdark model for each science corrtag.

    Args:
        corrtags (list or array-like): Predict the model superdark for these corrtags.
        superdarks (list or array-like): Superdarks to use for modeling. 
        segment (str): (Optional) Process corrtags of this segment only.
        hv (str): (Optional) Process corrtags of this HV only.
        outdir (str): (Optional) Output directory to write products and plots.
            Default is current working directory.
        binned (Bool): (Optional) If True, indicates input superdarks are already
            binned. If False, input superdarks will be binned on the fly.
        overwrite (Bool): If True, overwrite any pre-existing products.
    """
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    
    # Ensure all input superdarks are binned identically 
    check_superdarks(superdarks)

    for item in corrtags:
        do_interpolate_gsag, do_interpolate_lp1 = measure_gsag(item)
        if do_interpolate_gsag is True:
            break

    # If superdarks are not yet binned, bin them.
    if binned is False:
        binned_superdarks = []
        for darkfile in superdarks:
            S = Superdark.from_asdf(darkfile, overwrite=overwrite)
            binnedname0 = darkfile.replace(".asdf", "_binned.asdf")
            binnedname = os.path.join(outdir, binnedname0)
            binned_superdarks.append(binnedname)
            S.screen_counts(verbose=False, interpolate=True)
            S.bin_superdark(RESEL[0]*2, RESEL[1]*2, pha_bins=PHA_INCL_EXCL,
                             writefile=False)
            S.screen_counts(verbose=False, sigma=5, interpolate=True)
            if S.fixed_gsag is False and do_interpolate_gsag is True:
                S.fix_gsag(lp1_interpolate=do_interpolate_lp1)
            S.write_superdark(user_outfile=binnedname)
            S.plot_superdarks()
    else:
        binned_superdarks = superdarks
        for darkfile in superdarks:
            S = Superdark.from_asdf(darkfile, overwrite=True)
            S.screen_counts(verbose=False, sigma=5, interpolate=True)
            if S.fixed_gsag is False and do_interpolate_gsag is True:
                S.fix_gsag(lp1_interpolate=do_interpolate_lp1)
            S.write_superdark(user_outfile=darkfile)
            S.plot_superdarks()

    # Plot the binned superdarks 
    pdfname = os.path.join(outdir, "all_binned_superdarks.pdf")
    pdf = PdfPages(pdfname)
    dark_ims = []
    superdark_exptimes = []
    for darkfile in binned_superdarks:
        dark_af = asdf.open(darkfile)
        superdark_exptimes.append(dark_af["total_exptime"])
        binning = get_binning_pars(dark_af)
        pha_str = f"pha{PHA_INCL_EXCL[0]}-{PHA_INCL_EXCL[1]}"
        dark = dark_af[pha_str]
        this = type(dark.data)
        dark_im = copy.deepcopy(dark)
        dark_ims.append(dark_im)
        dark_hv = dark_af["hv"]
        dark_segment = dark_af["segment"]
        fig,ax = plt.subplots(1, 1, figsize=(20,8))
        im = ax.imshow(dark, aspect="auto", origin="lower", interpolation="nearest")
        fig.colorbar(im, label="Counts", pad=0.01)
        ax.set_title(f"{os.path.basename(darkfile)}\n{dark_segment} {dark_hv} superdark", size=25)
        pdf.savefig(fig, bbox_inches="tight")
        dark_af.close()
    pdf.close()
    print(f"Wrote {pdfname}")

    # Now go through each input science corrtag
    for item in corrtags:
        file_segment = fits.getval(item, "segment")
        file_hv = fits.getval(item, f"HVLEVEL{file_segment[-1]}", 1)
        if segment is not None:
            if file_segment != segment.upper() and str(file_hv) != str(hv):
                print(f"File does not match required settings: {item}")
                continue
        else:
            segment = file_segment
        if hv is not None:
            if file_hv != hv:
                print(f"File does not match required settings: {item}")
                continue

        cenwave = fits.getval(item, "cenwave")
        fig, ax = plt.subplots(1, 1, figsize=(20, 8))
        lp = fits.getval(item, "life_adj")
        excluded_rows, apertures = get_psa_wca(segment, cenwave, lp, binning)
        rootname0 = fits.getval(item, "rootname")
        rootname = rootname0.lower()
        binned_sci, nevents = bin_science(item, binning, segment, cenwave, lp, exclude_airglow=exclude_airglow)
#        vmin = np.mean(binned_sci) - 1*np.std(binned_sci)
#        if vmin < 0:
#            vmin = 0
        vmin = 0
        vmax = np.median(binned_sci) + np.median(binned_sci)*2.
        if vmax < 6:
            vmax = 6
        sh = np.shape(binned_sci)
        im = ax.imshow(binned_sci, aspect="auto", origin="lower", 
                vmin=vmin, vmax=vmax, extent=[0, sh[1], 0, sh[0]], interpolation="nearest")
        ax.axhline(apertures["PSA"][0], lw=2, color=COLORS[3], label="PSA")
        ax.axhline(apertures["PSA"][-1]+1, lw=2, color=COLORS[3])
        ax.axhline(apertures["WCA"][0], lw=2, ls="dashed", color=COLORS[3], label="WCA")
        ax.axhline(apertures["WCA"][-1]+1, lw=2, ls="dashed", color=COLORS[3])
        ax.legend(loc="upper right")
        fig.colorbar(im, label="Counts", pad=0.01)
        figname = os.path.join(outdir, f"{rootname}_{segment}_binned_sci.png")
        ax.set_title(f"{rootname} {segment} Binned Science Image", size=25)
        ax.set_xlabel("X (binned)")
        ax.set_ylabel("Y (binned)")
        exists = check_existing(figname, overwrite)
        if not exists:
            fig.savefig(figname, bbox_inches="tight")
            print(f"Saved binned science image: {figname}")
        plt.clf()
        sci_exp = fits.getval(item, "exptime", 1)

        nrows = np.shape(binned_sci)[0]
        all_rows = np.arange(nrows)
        bkg_rows = list(set(all_rows) - set(excluded_rows))

        # This plot is not really that useful nowadays.
#        # Plot the counts in each row that are used for scaling-
#        # that is, all non-PSA and non-WCA rows.        
#        fig, ax = plt.subplots(1, 1, figsize=(20, 8))
#        for i in range(binned_sci.shape[0]):
#            if i not in excluded_rows:
#                ax.plot(binned_sci[i], label=f"Row={i}", color=COLORS[i], alpha=0.8)
#        ax.set_title(f"{rootname}- all rows that are not in PSA/WCA", size=25)
#        ax.set_ylim(-0.5, 4)
#        ax.legend(loc="upper right")
#        figname = os.path.join(outdir, f"{rootname}_{segment}_rows.png")
#        exists = check_existing(figname, overwrite)
#        if not exists:
#            fig.savefig(figname, bbox_inches="tight")
#            print(f"Saved non-PSA/WCA rows: {figname}")
#        plt.clf()

        # Make an initial guess at the combined superdark which is just an equal
        # contribution from each binned superdark.
        equal = 1 / len(binned_superdarks)
        superdark_coeffs = [equal / (exptime / sci_exp) for exptime in superdark_exptimes]
        combined_dark = linear_combination(dark_ims, superdark_coeffs)

        # Now actually determine the best combination of all superdarks 
# TO DO, always hardcode this?
# Compare hardcoded vs variables
        #x0 = [0.007, 0.005]
        x0 = [0.1 * np.mean(binned_sci) / np.mean(dark_im) for dark_im in dark_ims]
#        x0 = superdark_coeffs
        res = minimize(fun_opt, x0, method="Nelder-Mead", tol=1e-6, 
                       args=(dark_ims, binned_sci, excluded_rows),
                       #bounds=[(1.e-8, None), (1.e-8, None)], options={'maxiter': 1000})
                       bounds=[(1.e-10, None) for x in binned_superdarks], options={'maxiter': 500})
        combined_dark1 = linear_combination(dark_ims, res.x)
        #print(res.x, lo_af["total_exptime"], hi_af["total_exptime"], sci_exp)
        now2 = datetime.datetime.now()
        print(now2)

        # Plot the model superdark
        sh = np.shape(combined_dark1)
        vmin = np.median(combined_dark1) - np.median(combined_dark1)*0.5
        if vmin < 0:
            vmin = 0
        vmax = np.median(combined_dark1) + np.median(combined_dark1)*2.
        fig, ax = plt.subplots(1, 1, figsize=(20, 8))
        im = ax.imshow(combined_dark1, aspect="auto", origin="lower",
                   extent=[0, sh[1], 0, sh[0]], interpolation="nearest",
                   vmin=vmin, vmax=vmax)
        fig.colorbar(im, label="Counts", pad=0.01)
        ax.set_title(f"{rootname} Best Model Superdark", size=25)
        ax.set_xlabel("X (binned)")
        ax.set_ylabel("Y (binned)")
        figname = os.path.join(outdir, f"{rootname}_{segment}_combined_dark.png")
        exists = check_existing(figname, overwrite)
        if not exists:
            fig.savefig(figname, bbox_inches="tight")
            print(f"Saved combined dark image: {figname}")
        plt.clf()

        # Plot the predicted vs. actual dark rates for all non-PSA and non-WCA rows
        ncols = 3
        nrows = int(np.ceil(len(bkg_rows)/ncols))
        fig, axes0 = plt.subplots(nrows, ncols, figsize=(25, nrows*6))
        fig.subplots_adjust(hspace=0.3, wspace=0.15)
        axes = axes0.flatten()
        for i in range(len(bkg_rows)):
            row = bkg_rows[i]
            ax = axes[i]
            smoothy, smoothx = smooth_array(binned_sci[row], 25)
            avg = np.average(binned_sci[row])
            avg_smoothy = np.average(smoothy)
            diff = np.abs(avg_smoothy - np.average(combined_dark1[row]))
            if diff > 2*avg_smoothy:
                ax.annotate("WARNING!\nBAD FIT!", (.05, .85), color=COLORS[2], xycoords="axes fraction")    
            ax.plot(binned_sci[row], color="lightgrey", 
                    label=f"Science", alpha=0.8)
            ax.plot(combined_dark1[row], color=COLORS[0], 
                    label=f"Predicted Bkgd.", alpha=0.8)
            ax.plot(smoothx, smoothy, color=COLORS[1], label="Smoothed Sci.", alpha=0.8)
            smoothy_max = np.max(smoothy) * 2.
            ylim = ax.get_ylim()
            if ylim[1] > smoothy_max:
                ax.set_ylim(top=smoothy_max)
#            ax.axhline(avg, lw=2, color=COLORS[2], label=f"Avg Bkgd={avg:.1f}")
            ax.set_title(f"Binned row = {row}", size=18)
            ax.set_xlabel("X (binned)")
            ax.set_ylabel("Counts")
        axes[2].legend(loc="upper right", fontsize=18) 
        #handles, labels = ax.get_legend_handles_labels()
        #fig.legend(handles, labels, bbox_to_anchor=(.9, .91), loc="lower right", fontsize=18) 
        fig.suptitle(f"Predicted vs. actual dark for non-PSA/WCA rows\n{rootname} {segment}", size=25, y=.95)
        figname = os.path.join(outdir, f"{rootname}_{segment}_predicted_dark_bkgd.png")
        exists = check_existing(figname, overwrite)
        if not exists:
            plt.savefig(figname, bbox_inches="tight")
            print(f"Saved actual vs predicted darks: {figname}")
        plt.clf()
        
        # Plot the predicted vs. actual dark rates for all PSA rows
        ncols = 1
        nrows = int(np.ceil(len(apertures["PSA"])/ncols))
        fig, axes0 = plt.subplots(nrows, ncols, figsize=(22, nrows*6))
        fig.subplots_adjust(hspace=0.3)
        axes = axes0.flatten()
        for i in range(len(apertures["PSA"])):
            row = apertures["PSA"][i]
            ax = axes[i]
            smoothy, smoothx = smooth_array(binned_sci[row], 25)
            avg = np.average(binned_sci[row])
            ax.plot(binned_sci[row], label="Science", color="lightgrey", alpha=0.8)
            ax.plot(combined_dark1[row], label=f"Predicted Bkgd.", color=COLORS[0], alpha=0.8)
            ax.plot(smoothx, smoothy, color=COLORS[1], label="Smoothed Sci.", alpha=0.8)
#            ax.axhline(avg, lw=2, color=COLORS[2], label=f"Avg Sci={avg:.1f}")
            ax.set_title(f"Binned row = {row}", size=20)
            ax.set_xlabel("X (binned)")
            ax.set_ylabel("Counts")
            ax.set_ylim(bottom=-.75)
            smoothy_max = np.max(smoothy) * 2.
            ylim = ax.get_ylim()
            if ylim[1] > smoothy_max:
                ax.set_ylim(top=smoothy_max)
        axes[0].legend(loc="upper right", fontsize=18) 
        #handles, labels = ax.get_legend_handles_labels()
        #fig.legend(handles, labels, bbox_to_anchor=(.9, .91), loc="lower right", fontsize=18) 
        figname = os.path.join(outdir, f"{rootname}_{segment}_predicted_dark_sci.png")
        fig.suptitle(f"Predicted vs. actual dark across PSA rows\n{rootname} {segment}", size=25, y=.97)
        exists = check_existing(figname, overwrite)
        if not exists:
            plt.savefig(figname, bbox_inches="tight")
            print(f"Saved predicted dark vs science: {figname}")
        plt.clf()

#       These two files below are used for sanity checks
#        noise = copy.deepcopy(lo_af.tree)
#        noise[pha_str] = combined_dark1[apertures["PSA"]]
#        noise["scaling"] = res.x
#        outfile = os.path.join(outdir, f"{rootname}_{segment}_noise.asdf")
#        noise_af = asdf.AsdfFile(noise)
#        exists = check_existing(outfile, overwrite)
#        if not exists:
#            noise_af.write_to(outfile)
#            print(f"Wrote {outfile}")
#
#        signal = copy.deepcopy(lo_af.tree)
#        signal[pha_str] = binned_sci[apertures["PSA"]]
#        outfile = os.path.join(outdir, f"{rootname}_{segment}_signal.asdf")
#        signal_af = asdf.AsdfFile(signal)
#        exists = check_existing(outfile, overwrite)
#        if not exists:
#            signal_af.write_to(outfile)
#            print(f"Wrote {outfile}")
        
        dark_af = asdf.open(binned_superdarks[0])
        noise_comp = copy.deepcopy(dark_af.tree)
        noise_comp[pha_str] = combined_dark1
        #combined_exptime = (lo_exptime * res.x[0]) + (hi_exptime * res.x[1])
        noise_comp["total_exptime"] = sci_exp
        outfile = os.path.join(outdir, f"{rootname}_{segment}_noise_complete.asdf")
        noise_comp_af = asdf.AsdfFile(noise_comp)
        exists = check_existing(outfile, overwrite)
        if not exists:
            noise_comp_af.write_to(outfile)
            print(f"Wrote {outfile}")
        dark_af.close()




import os
import itertools
import numpy as np
import argparse
from matplotlib import pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "niceplot.mplstyle")
plt.style.use(stylesheet)
import glob
from astropy.io import fits
from scipy import stats
import asdf

from acdc.superdark.cos_fuv_superdark import Superdark
from acdc.utils.utils import get_binning_pars, bin_coords

LOCAL_REFDIR = "/grp/hst/cdbs/lref"
LOCAL_DARKDIR = "/astro/sveash/cos_dark/final_superdarks"
PHA_INCLUSIVE = [2, 23]
PHA_INCL_EXCL = [2, 24]
# Colorblind-safe palette below
COLORS = ["#1b9e77", "#d95f02", "#7570b3", "#e7298a", "#66a61e", "#a6cee3",
          "#1f78b4", "#b2df8a", "#33a02c", "#fb9a99", "#e31a1c", "#fdbf6f",
          "#ff7f00", "#cab2d6", "#6a3d9a", "#ffff99", "#b15928", "#a6cee3"]


def sum_bkg_region(data, bkg_lo, bkg_hi):
    assert np.shape(data)[1] == len(bkg_lo) == len(bkg_hi), "Data X-dimension must match limits of background regions"
    datasum = []
    for i in range(len(bkg_lo)):
        colsum = np.sum(data[bkg_lo[i]:bkg_hi[i], i])
        datasum.append(colsum)
    datasum = np.array(datasum)
    
    return datasum

def smooth_array(data, smoothsize=100):
    bin_edges = np.arange(0, len(data), smoothsize)
    binned_y, edges, binnumber = stats.binned_statistic(np.arange(len(data)),
                                                      data,
                                                      statistic="mean",
                                                      bins=bin_edges)
    binned_x = bin_edges[:-1] + (smoothsize/2.)
    return binned_y, binned_x

def get_data(flt_custom, flt_default, superdark_custom, dark_dir=LOCAL_DARKDIR, ref_dir=LOCAL_REFDIR):
    data_custom = fits.getdata(flt_custom)
    data_default = fits.getdata(flt_default)
    segment = fits.getval(flt_custom, "segment")
    cenwave = fits.getval(flt_custom, "cenwave")
    hv = fits.getval(flt_custom, f"HVLEVEL{segment[-1]}", 1)
    xtractab_file0 = fits.getval(flt_custom, "xtractab")
    xtractab_file1 = xtractab_file0.split("$")[1]
    xtractab_file = os.path.join(ref_dir, xtractab_file1)

    xtract = fits.getdata(xtractab_file)
    inds = np.where((xtract["segment"] == segment) &
                    (xtract["cenwave"] == cenwave) & 
                    (xtract["aperture"] == "PSA"))
    xtract_data = xtract[inds]
    b1 = xtract_data["b_bkg1"]
    b2 = xtract_data["b_bkg2"]
    height = xtract_data["height"]
    m = xtract_data["slope"]
    smoothsize = xtract_data["bwidth"]
    x = np.arange(0, 16384)
    bkg1_mid = m*x + b1
    bkg1_y1 = bkg1_mid + height/2
    bkg1_y0 = bkg1_mid - height/2
    bkg2_mid = m*x + b2
    bkg2_y1 = bkg2_mid + height/2
    bkg2_y0 = bkg2_mid - height/2
    bkg1_y0 = np.round(bkg1_y0)
    bkg1_y0 = bkg1_y0.astype(int)
    bkg2_y0 = np.round(bkg2_y0)
    bkg2_y0 = bkg2_y0.astype(int)
    bkg1_y1 = np.round(bkg1_y1)
    bkg1_y1 = bkg1_y1.astype(int)
    bkg2_y1 = np.round(bkg2_y1)
    bkg2_y1 = bkg2_y1.astype(int)

    superdark_lo = glob.glob(os.path.join(dark_dir, f"*{segment}*{hv}*quiescent*asdf"))
    superdark_lo = [x for x in superdark_lo if "binned" not in x]
    superdark_hi = glob.glob(os.path.join(dark_dir, f"*{segment}*{hv}*active*asdf"))
    superdark_hi = [x for x in superdark_hi if "binned" not in x]
    assert len(superdark_lo) != 0, \
        f"No quiescent superdarks found for {segment} {hv}" 
    assert len(superdark_hi) != 0, \
        f"No active superdarks found for {segment} {hv}" 
    assert len(superdark_lo) == 1, \
        f"Too many quiescent superdarks found for {segment} {hv}" 
    assert len(superdark_hi) == 1, \
        f"Too many active superdarks found for {segment} {hv}"
    superdark_lo = superdark_lo[0]
    superdark_hi = superdark_hi[0]
    S_lo = Superdark.from_asdf(superdark_lo) 
    S_lo.bin_superdark(1, 1, pha_bins=[1,31]) 
    #S_lo.bin_superdark(1, 1, pha_bins=PHA_INCL_EXCL) 
    S_hi = Superdark.from_asdf(superdark_hi) 
    S_hi.bin_superdark(1, 1, pha_bins=[1,31])
    #S_hi.bin_superdark(1, 1, pha_bins=PHA_INCL_EXCL)
    lo_exptime = S_lo.total_exptime
    hi_exptime = S_hi.total_exptime
    S_lo_flt0 = S_lo.superdarks[0] / lo_exptime
    S_hi_flt0 = S_hi.superdarks[0] / hi_exptime
    assert np.shape(S_lo_flt0) == np.shape(S_hi_flt0), \
        "Quiescent and active superdark dimensions do not match"
    xstart = S_lo.bin_xstart
    xend = S_lo.bin_xend
    ystart = S_lo.bin_ystart
    yend = S_lo.bin_yend
    S_lo_flt = np.zeros(16777216).reshape(1024, 16384)
    S_hi_flt = np.zeros(16777216).reshape(1024, 16384)
    S_lo_flt[ystart:yend, xstart:xend] = S_lo_flt0
    S_hi_flt[ystart:yend, xstart:xend] = S_hi_flt0

    bkg1_sum_lo = sum_bkg_region(S_lo_flt, bkg1_y0, bkg1_y1)
    bkg1_lo,_ = smooth_array(bkg1_sum_lo)
    bkg2_sum_lo = sum_bkg_region(S_lo_flt, bkg2_y0, bkg2_y1)
    bkg2_lo,_ = smooth_array(bkg2_sum_lo)
    bkg1_sum_hi = sum_bkg_region(S_hi_flt, bkg1_y0, bkg1_y1)
    bkg1_hi,_ = smooth_array(bkg1_sum_hi)
    bkg2_sum_hi = sum_bkg_region(S_hi_flt, bkg2_y0, bkg2_y1)
    bkg2_hi,_ = smooth_array(bkg2_sum_hi)

    bkg1_sum_custom = sum_bkg_region(data_custom, bkg1_y0, bkg1_y1)
    bkg1_custom,_ = smooth_array(bkg1_sum_custom)
    bkg2_sum_custom = sum_bkg_region(data_custom, bkg2_y0, bkg2_y1)
    bkg2_custom,_ = smooth_array(bkg2_sum_custom)
    bkg1_sum_default = sum_bkg_region(data_default, bkg1_y0, bkg1_y1)
    bkg1_default,_ = smooth_array(bkg1_sum_default)
    bkg2_sum_default = sum_bkg_region(data_default, bkg2_y0, bkg2_y1)
    bkg2_default, binned_x = smooth_array(bkg2_sum_default)
    
    af = asdf.open(superdark_custom)
    pha_str = f"pha{PHA_INCL_EXCL[0]}-{PHA_INCL_EXCL[1]}"
    dark_im = af.tree[pha_str]
    dark_im_flt0 = dark_im / af.tree["total_exptime"]
    binning = get_binning_pars(af)
    dark_im_flt_perpixel = dark_im_flt0 / binning["bin_y"] / binning["bin_x"]
    dark_im_flt = np.zeros(16777216).reshape(1024, 16384)
    xs = np.arange(binning["xstart"], binning["xend"], binning["bin_x"])
    ys = np.arange(binning["ystart"], binning["yend"], binning["bin_y"])
    for i in range(len(xs)-1): 
        for j in range(len(ys)-1):
            dark_im_flt[ys[j]:ys[j+1], xs[i]:xs[i+1]] = dark_im_flt_perpixel[j,i]
    dark_x = np.arange(16384)
    
    bkg1_dark = sum_bkg_region(dark_im_flt, bkg1_y0, bkg1_y1)
    bkg2_dark = sum_bkg_region(dark_im_flt, bkg2_y0, bkg2_y1)
    
    return binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, \
        bkg1_lo, bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x
    #return np.arange(16384), bkg1_sum_default, bkg2_sum_default, bkg1_sum_custom, bkg2_sum_custom, bkg1_sum_lo, bkg2_sum_lo, bkg1_sum_hi, bkg2_sum_hi


def plot_data(x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, bkg1_lo, 
              bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x, targname):
    fig, axes0 = plt.subplots(3, 1, figsize=(8.5, 11))
    axes = axes0.flatten()
    axes[0].hlines(y=0, xmin=-1000, xmax=17384, colors="lightgrey", linestyles="--")
    axes[1].hlines(y=0, xmin=-1000, xmax=17384, colors="lightgrey", linestyles="--")
    axes[2].hlines(y=0, xmin=-1000, xmax=17384, colors="lightgrey", linestyles="--")
    
    bkg1 = {"default": bkg1_default, "custom": bkg1_custom, "lo": bkg1_lo, "hi": bkg1_hi, "superdark": bkg1_dark}
    bkg2 = {"default": bkg2_default, "custom": bkg2_custom, "lo": bkg2_lo, "hi": bkg2_hi, "superdark": bkg2_dark}
    av = {"default": (bkg1_default + bkg2_default) / 2.,
          "custom": (bkg1_custom + bkg2_custom) / 2.,
          "lo": (bkg1_lo + bkg2_lo) / 2.,
          "hi": (bkg1_hi + bkg2_hi) / 2.,
          "superdark": (bkg1_dark + bkg2_dark) / 2.}
    dicts = [bkg1, bkg2, av]

    titles = ["Background 1 (Lower)", "Background 2 (Upper)", "Average"]
    for i in range(3):
        ax = axes[i]
        d = dicts[i]
        ax.set_ylabel("Counts/s")
        ax.plot(x, d["default"],   COLORS[0], alpha=0.8, label="Default FLT")
        ax.plot(x, d["custom"],    COLORS[1], alpha=0.8, label="Custom FLT")
        ax.plot(x, d["lo"],        COLORS[2], alpha=0.8, label="Lo Dark")
        ax.plot(x, d["hi"],        COLORS[3], alpha=0.8, label="Hi Dark")
        ax.plot(dark_x, d["superdark"], COLORS[4], alpha=0.8, label="Custom Dark")
        ax.set_ylim(-.0001, 0.00035)
        ax.set_xlim(-500, 16884)
        ax.set_title(titles[i])
        ax.legend()
    figname = f"{targname}_flt_comp.png"
    fig.savefig(figname)
    print(f"Wrote {figname}")


def compare_backgrounds(flt_default, flt_custom, targname, superdark_custom):
    binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, bkg1_lo, \
        bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x = get_data(flt_custom, flt_default, superdark_custom)
    plot_data(binned_x, bkg1_default, bkg2_default, bkg1_custom, bkg2_custom, \
        bkg1_lo, bkg2_lo, bkg1_hi, bkg2_hi, bkg1_dark, bkg2_dark, dark_x, targname)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--default",
                        help="Name of default CalCOS FLT file")
    parser.add_argument("-c", "--custom",
                        help="Name of custom-corrected FLT file")
    parser.add_argument("-t", "--targname",
                        help="Name of target")
    parser.add_argument("-s", "--superdark_custom",
                        help="Name of combined superdark from ACDC")
    args = parser.parse_args()
    compare_backgrounds(args.default, args.custom, args.targname, args.superdark_custom)




"""
Perform a custom dark correction on COS/FUV data.
For each input science exposure, a quiescent and an active superdark 
of the appropriate segment+HV combination are used to determine the dark rate 
of the science exposure. A custom dark correction is then applied using
this model dark rate, creating custom corrtag products. These corrtags are
calibrated using CalCOS to create custom x1d files. x1d files should be coadded
offline in order to increase SNR.
"""

import os
DIRNAME = os.path.dirname(__file__)
import sys
import gzip
import shutil
from collections import defaultdict
import argparse
import glob
import datetime
from astropy.io import fits
import asdf
import calcos

from acdc.predict_dark_level import predict_dark
from acdc.subtract_dark_level import subtract_dark


class Acdc():
    """Perform a custom dark correction on COS/FUV data.

    Attributes:
        indir (str): Input directory that houses corrtags to correct.
        darkcorr_outdir (str): Custom dark-corrected corrtags will be written here.
        x1d_outdir (str): Custom dark-corrected x1ds will be written here.
        binned (Bool): If True, pre-binned superdarks will be used.
        corr_dict (dict): Dictionary where each key is the segment+HV setting 
            and each value is a list of all corrtags with that setting.
        custom_corrtags (list): List of corrtags which have had the custom
            dark correction applied.
        dark_dict (dict): Nested dictionary where each key is the segment+HV setting
            and each value is a dictionary where each key is the type of superdark 
            (either 'active' or 'quiescent') and each value is the superdark name.
    """
    
    def __init__(self, indir, darkcorr_outdir, x1d_outdir=None, binned=True, 
                 segment=None, hv=None, overwrite=False,
                 exclude_airglow=False, superdark_dir=None, 
                 superdarks=None, calibrate=True):
        """
        Args:
            indir (str): Input directory that houses corrtags to correct.
            darkcorr_outdir (str): Custom dark-corrected corrtags will be written here.
            x1d_outdir (str): Custom dark-corrected x1ds will be written here.
            binned (Bool): If True, pre-binned superdarks will be used.
            superdark_dir (str): Location of superdarks. 
        """

        self.calibrate = calibrate
        self.overwrite = overwrite
        self.indir = indir
        self.exclude_airglow = exclude_airglow
        if superdark_dir is None and superdarks is None:
            trydir = os.path.join(DIRNAME, "data/superdarks/epoch_superdarks")
            if len(glob.glob(os.path.join(trydir, "*superdark*asdf*"))) != 0:
                superdark_dir = trydir 
            else:
                try:
                    superdark_dir = os.environ["ACDC_SUPERDARKS"]
                except KeyError as e:
                    print("ERROR: You must supply the superdark directory or define the $ACDC_SUPERDARKS environment variable- this is where all superdarks are located")
                    print("Exiting")
                    sys.exit()
        self.superdark_dir = superdark_dir
        if superdarks is None:
            superdarks = glob.glob(os.path.join(self.superdark_dir, "*superdark*.asdf*"))
#        superdarks = self.unzip_superdarks(superdarks)
        
        self.darkcorr_outdir = darkcorr_outdir
        self.binned = binned
        if segment is not None:
            segment = segment.upper()
        self.segment = segment 
        if hv is not None:
            hv = int(hv)
        self.hv = hv 
        now = datetime.datetime.now()
        self.x1d_outdir = os.path.join(darkcorr_outdir, f"cal_{now.strftime('%d%b%Y')}")
        if not os.path.exists(darkcorr_outdir):
            os.makedirs(darkcorr_outdir)
        corrtags = glob.glob(os.path.join(indir, "*corrtag*fits"))
        self.corr_dict = self.sort_corrtags(corrtags)
        self.dark_dict = self.get_best_superdarks(superdarks)

    
    def sort_corrtags(self, corrtags):
        """
        Sort corrtags into a dictionary. If specified, filter corrtags based 
        on segment and HV setting.
        
        Args:
            corrtags (array-like): All corrtags to be corrected.
        
        Returns:
            corr_dict (dict): Dictionary where each key is the segment+HV setting 
                and each value is a list of all corrtags with that setting.
        """
        
        corr_dict = defaultdict(list)
        for item in corrtags:
            file_segment = fits.getval(item, "segment")
            file_hv = fits.getval(item, f"HVLEVEL{file_segment[-1]}", 1)
            if self.segment is not None:
                if file_segment != self.segment:
                    continue
            if self.hv is not None:
                if int(file_hv) != self.hv:
                    continue
            corr_dict[f"{file_segment}_{file_hv}"].append(item)
        assert len(corr_dict) != 0, "No corrtags found that match segment and HV constraints"
        return corr_dict

    
    def get_best_superdarks(self, superdarks):
        """Sort superdarks into a dictionary based on segment and HV setting.
        
        Returns:
            dark_dict (dict): Dictionary where each key is the segment+HV setting
                and each value is a list of all applicable superdarks. 
        """
        if self.binned is False:
            darks = [x for x in superdarks if "binned" not in x]
        else:
            darks = [x for x in superdarks if "binned" in x]
        dark_dict = defaultdict(list)
        corr_segments = [x.split("_")[0] for x in self.corr_dict]
        corr_hvs = [x.split("_")[1] for x in self.corr_dict]
        for dark in darks:
            darkfile = os.path.basename(dark)
            sp = darkfile.split("_")
            segment = sp[1]
            hv = sp[2]
            for i in range(len(corr_hvs)):
                if corr_segments[i] == segment and corr_hvs[i] == hv:
                    dark_dict[f"{segment}_{hv}"].append(dark)
        
        assert len(dark_dict) != 0, "No matching superdarks found!!"
        if self.binned is False:
            print("Matching unbinned superdarks:")
        else:
            print("Matching binned superdarks:")
        for k,v in dark_dict.items():
            for sd in v:
                print(f"\t{sd}")

        return dark_dict

    
    def unzip_superdarks(self, superdarks):
        unzipped_superdarks = []
        for item in superdarks:
            if item.endswith(".gz"):
                unzipped = item.replace(".gz", "")
                with gzip.open(item, 'rb') as f_in:
                    with open(unzipped, 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)
                print(f"Unzipped {item}")
                unzipped_superdarks.append(unzipped)
            else:
                unzipped_superdarks.append(item)
        return unzipped_superdarks
   

    def custom_dark_correction(self):
        """Perform the custom dark correction.

        For each corrtag, use the superdarks to predict the dark rate across 
        the entire detector. Then use this predicted dark rate to perform a 
        custom dark correction to each corrtag.
        """
        
        for seg_hv in self.corr_dict:
            corrtags = self.corr_dict[seg_hv]
            superdarks = self.dark_dict[seg_hv]
            predict_dark(corrtags, superdarks, 
                         outdir=self.darkcorr_outdir, binned=self.binned,
                         overwrite=self.overwrite, segment=self.segment,
                         hv=self.hv, exclude_airglow=self.exclude_airglow)
            subtract_dark(corrtags, self.darkcorr_outdir, outdir=self.darkcorr_outdir, 
                          overwrite=self.overwrite)
        self.custom_corrtags = glob.glob(os.path.join(self.darkcorr_outdir, "corrected*corrtag*fits"))

    
    def calibrate_corrtags(self):
        """Calibrate custom dark-corrected corrtags to product x1ds.

        Calibrate each custom dark-corrected corrtag using the BOXCAR extraction
        method in CalCOS. The TWOZONE extraction method does not work with
        BACKCORR=OMIT.
        """

        if self.calibrate is False:
            print("Calibration is set to skip, ending ACDC now")
            return

        for item in self.custom_corrtags:
            with fits.open(item, mode="update") as hdulist:
                hdr0 = hdulist[0].header
                hdr0.set("xtrctalg", "BOXCAR")
                hdr0.set("backcorr", "OMIT")
                hdr0.set("trcecorr", "OMIT")
                hdr0.set("algncorr", "OMIT")

        calibrated = []
        for item in self.custom_corrtags:
            if "_corrtag_a" in item:
                other = item.replace("_corrtag_a", "_corrtag_b")
            else:
                other = item.replace("_corrtag_b", "_corrtag_a")
            if other in calibrated:
                continue
            spl = os.path.basename(item).split("_")
            wildcard = spl[0] + "_" + spl[1] + "*"
            wildcard_products = glob.glob(os.path.join(self.x1d_outdir, wildcard))
            if len(wildcard_products) >= 1 and self.overwrite is True:
                for prod in wildcard_products:
                    os.remove(prod)
            elif len(wildcard_products) >= 1 and self.overwrite is False:
                print(f"WARNING: Products already exist and overwrite is False, skipping...")
                continue

            calcos.calcos(item, outdir=self.x1d_outdir, verbosity=0)
            calibrated.append(item)
            calibrated.append(other)




import sqlite3
from sqlite3 import Error
import sys
from math import * 
import numpy as np
from scipy.optimize import minimize
from scipy.special import factorial
import matplotlib as mpl
import matplotlib.pyplot as plt
plt.rc('font', family='serif')
mpl.rcParams.update({'font.size': 12})
mpl.rcParams.update({'legend.labelspacing':0.25, 'legend.fontsize': 12})
mpl.rcParams.update({'errorbar.capsize': 4})

## Connecting to the database
def create_connection(db_file):
	""" create a database connection to a SQLite database """
	conn = None
	try:
		conn = sqlite3.connect(db_file)
		#print(sqlite3.version)
	except Error as e:
		print(e)
	return conn

def select_all_tasks(conn, hv, segment, idl, idr):
	cur = conn.cursor()
	cur.execute("SELECT id, latitude, longitude, expstart, region, solar_flux, dark_pha0, dark_pha1, dark_pha2, dark_pha3, \
                 dark_pha4,  dark_pha5,  dark_pha6,  dark_pha7,  dark_pha8,  dark_pha9,  dark_pha10, dark_pha11, \
                 dark_pha12, dark_pha13, dark_pha14, dark_pha15, dark_pha16, dark_pha17, dark_pha18, dark_pha19, \
                 dark_pha20, dark_pha21, dark_pha22, dark_pha23, dark_pha24, dark_pha25, dark_pha26, dark_pha27, \
                 dark_pha28, dark_pha29, dark_pha30, dark_pha31 \
                 FROM Darks  \
                 WHERE region='inner' and segment='"+str(segment)+"'  and hv="+str(hv)+' and id >'+str(idl)+' and id <'+str(idr))


	rows = cur.fetchall()

	data = np.zeros(32)

	print ('Number of rows: ', len(rows))

	sf = []
	exp_time = []
	cn = 0

	for row in rows:

		#print ('---> ', row[3])

		if cn > 0:
	
			if abs(float(row[3]) - cmp_str) > 0.5:
				break

		else:
			cmp_str = float(row[3])
			cn = cn + 1

		sf.append (float(row[5]))
		exp_time.append (float(row[3]))
		

		for i in range (0, 32):
	
			data[i] = data[i] + float(row[i+6])
			
		#break

	#print ('MEAN::: ', np.mean(sf))

	if np.max(data) == 0:
		sf.append(0)
		exp_time.append(0)

	#print ('Dates: ', exp_time)

	return [data, np.mean(sf), np.mean(exp_time)]



## Exponential distribution
def exp_fun (E, k, Et):

	val = k * np.exp (-E / Et) #/ (Et *(1.0 - exp(-31.0/Et)) )

	return val

## Exponential + normal distribution
def exp_norm_fun (E, k, Et, w, mu, sigma):

	val = k *np.exp (-E / Et) + w * np.exp (-np.power(E - mu, 2.0) / 2./np.power(sigma,2.0)) #/ sqrt(2.*pi) / sigma

	return val

## Likelihood in form of C-statistics for exponential distribution
def Cstat (param, data):

	k, Et = param

	E = np.linspace (0,31, 32)


	model = exp_fun (E, k, Et)

	if np.min (data) == 0:
		C = 2 * (model - data * np.log(model) + np.log(factorial(data)))
	else:
		C = 2 * (model - data + data * (np.log(data) - np.log(model)))
#	for i in range (0, len(C)):
#		print (i, C[i])

#	print ('.. ', np.sum(C), np.sum(C[2:29]))

#	sys.exit(0)	

	C = np.sum(C[2:29])

	

	#print ('---', k, Et, C)

	return C

## Likelihood in form of C-statistics for exponential + normal distribution
def Cstat_alt (param, data):

	k, Et, w, mu, sigma = param


	E = np.linspace (0,31, 32)

	model = exp_norm_fun (E, k, Et, w, mu, sigma)

	if np.min (data) == 0:
		C = 2 * (model - data * np.log(model) + np.log(factorial(data)))
	else:
		C = 2 * (model - data + data * (np.log(data) - np.log(model)))

	C = np.sum(C[2:29])

	#print ('---', k, Et, w, mu, sigma, C)

	if isnan(C):

		C = 5e12
	


	return C


## Function which determines which of two models describe the spectra the best:
## (1) - exponential distribution, in this case the function returns two parameters: k, Et
## (2) - exponential + normal distribution, in this case the function returns five parameters: k, Et for exponential
## and w, mu and sigma for normal distribution

def study_spectra (data):

	x0 = (np.max(data[3:29]), 10)

	res = minimize(Cstat, x0, method='L-BFGS-B', tol=1e-6, args=(data), bounds=((0, 2.0*np.max(data)), (0.1, 60))   )

	Cstat_A = res.fun
	A_par   = res.x
	
	
	x0 = (A_par[0], A_par[1], A_par[0]/3.0, 10, 3)

	res = minimize(Cstat_alt, x0, method='L-BFGS-B', tol=1e-6, args=(data), bounds=((0, 2.0*np.max(data)), (0.1, 60), (0, np.max(data)), (3,31), (0.1,60)   ))

	Cstat_B = res.fun
	B_par   = res.x

	print ('Cstat for A: ', Cstat_A, ' Cstat for B: ', Cstat_B)

	if Cstat_B < Cstat_A - 2*3: ##  AIC - model B has three more parameters

		return B_par

	else:	
	
		return A_par


if __name__ == "__main__":

	hv = 167
	segment = 'FUVA'

	E = np.linspace (0,31, 32)
	conn = create_connection('../cos_dark.db')

	ii = []
	mu = []
	sf_l = []
	ff = []
	exp_date_ff = []

	ii_exp = []
	ff_exp = []
	exp_date_exp = []

	for i in range (0, 70000): ##70000


		data, sf, exp_date = select_all_tasks(conn, hv, segment, 100*i, 100*(i+1))

		print ('---', i, sf)

	
		if np.max(data) == 0:
			continue
	
		val = study_spectra (data)

		if len(val) < 3:

			model = exp_fun (E, val[0], val[1])

			ii_exp.append (i)
			ff_exp.append (0)
			exp_date_exp.append (exp_date)

		else:

			model = exp_norm_fun (E, val[0], val[1], val[2], val[3], val[4])

			ii.append (i)
			mu.append (val[3])
			sf_l.append (sf)
			ff.append (val[2] / val[0])
			exp_date_ff.append (exp_date)

		#print ('Result of optimisation: ',i,' model = ', val)
		#plt.plot (E, data, 'k-', drawstyle='steps-mid', label='Data')
		#plt.plot (E, model,       'b--',label='Model')
		#plt.xlabel ('E')
		#plt.ylabel ('Counts')
		#plt.legend()
	#plt.savefig ('fit_data.pdf')

	plt.scatter (exp_date_ff, mu, s=8)
	plt.xlabel('MJD')
	plt.ylabel(r'$\mu$')
	plt.ylim([2,27])
	plt.plot ([56012, 56012], [0, 30], 'k--')
	plt.plot ([56131, 56131], [0, 30], 'k--')
	plt.plot ([56467, 56467], [0, 30], 'k--')
	plt.plot ([56859, 56859], [0, 30], 'k--')
	plt.plot ([56964, 56964], [0, 30], 'k--')
	plt.plot ([57267, 57267], [0, 30], 'k--')
	plt.plot ([57405, 57405], [0, 30], 'k--')
	plt.plot ([57678, 57678], [0, 30], 'k--')
	plt.plot ([58028, 58028], [0, 30], 'k--')
	plt.savefig ('mu.pdf')
	plt.show()

	#print ('Sizes of rows: ', len(sf), len(mu))
	
	plt.scatter (sf_l, mu, s=8)
	plt.xlabel('Solar flux')
	plt.ylabel(r'$\mu$')	
	plt.savefig('solar_flux_mu.pdf')
	plt.show()

	print ('Correlation coeffitient: ',  np.corrcoef(sf_l, ff))

	plt.scatter (sf_l, ff, s=7)
	plt.xlabel('Solar flux')
	plt.ylabel(r'B/A')
	plt.ylim([0,5])
	plt.savefig('BA_solar_flux.pdf')
	plt.show()


	plt.scatter (exp_date_ff, ff, s=8)
	plt.scatter (exp_date_exp, ff_exp, color='red', s=8)
	plt.xlabel ('MJD')
	plt.ylabel (r'$B/A$')
	plt.ylim([0, 5])
	plt.plot ([56012, 56012], [0, 5], 'k--')
	plt.plot ([56131, 56131], [0, 5], 'k--')
	plt.plot ([56467, 56467], [0, 5], 'k--')
	plt.plot ([56859, 56859], [0, 5], 'k--')
	plt.plot ([56964, 56964], [0, 5], 'k--')
	plt.plot ([57267, 57267], [0, 5], 'k--')
	plt.plot ([57405, 57405], [0, 5], 'k--')
	plt.plot ([57678, 57678], [0, 5], 'k--')
	plt.plot ([58028, 58028], [0, 5], 'k--')



	plt.savefig ('BA.pdf')
	plt.show()

		#plt.show()

	#if len(val) < 3:

	#	model = exp_fun (E, val[0], val[1])

	#else:

	#	model = exp_norm_fun (E, val[0], val[1], val[2], val[3], val[4])


	#plt.plot (E, data, 'k-', drawstyle='steps-mid', label='Data')
	#plt.plot (E, model,       'b--',label='Model')
	#plt.xlabel ('E')
	#plt.ylabel ('Counts')
	#plt.legend()
	#plt.savefig ('fit_data.pdf')

	#plt.show()







import os
import argparse
from matplotlib.backends.backend_pdf import PdfPages
import copy
import asdf
import numpy as np
import matplotlib.pyplot as plt
dirname = os.path.dirname(__file__)
stylesheet = os.path.join(dirname, "../analysis", "niceplot.mplstyle")
plt.style.use(stylesheet)

def bin_superdark(superdark, bin_x=8, bin_y=2, bin_pha=26, 
                  phastart=3, phaend=28, outdir="."):

    with asdf.open(superdark) as af:
        # First bin by PHA

        binned = copy.deepcopy(af.tree[f"pha3-29"])
        binned_ims = {"3-29": binned}
        binned = copy.deepcopy(af.tree[f"pha{phastart}-{phastart+1}"])
        binned_ims = {}
        counter = 1
        for i in range(phastart+1, phaend+2):
            if counter == bin_pha:
                counter = 0
                binned_ims[f"{i-bin_pha}-{i}"] = binned
                binned = copy.deepcopy(af.tree[f"pha{i}-{i+1}"])
            else:
                binned += af.tree[f"pha{i}-{i+1}"]
            counter += 1

        sh = np.shape(binned)
        xdim = sh[1]
        ydim = sh[0]
        b_x0 = 0
        b_x1 = (xdim // bin_x) * bin_x
        b_y0 = 0
        b_y1 = (ydim // bin_y) * bin_y

        plt.imshow(binned_ims["3-29"], aspect="auto", origin="lower")
        plt.show()
        cont = input("press enter")

        pdffile = os.path.join(outdir, superdark.replace("asdf", "pdf"))
        pdf = PdfPages(pdffile)
        for k,binned in binned_ims.items():
            binned = binned[b_y0:b_y1, b_x0:b_x1]
            binned_sh = np.shape(binned)
            binned_xdim = binned_sh[1]
            binned_ydim = binned_sh[0]
            tmp = binned.reshape(binned_ydim // bin_y, bin_y, binned_xdim // bin_x, bin_x)
            binned = tmp.sum(axis=3).sum(axis=1)
            binned_ims[k] = binned
            rate = binned/af["total_exptime"]
            spl = k.split("-")
            print(f"For PHAs {spl[0]} through {spl[1]}:")
            print(f"\tTotal number of events: {np.sum(binned):,}")
            print(f"\tTotal exptime of superdark: {af['total_exptime']:,}")
            print(f"\tMinimum number of events in a binned pixel: {np.min(binned)}")
            print(f"\tMean number of events per binned pixel: {np.mean(binned):.1f}")
            print(f"\t  Standard deviation: {np.std(binned):.1f}")
            print(f"\tMedian number of events per binned pixel: {np.median(binned):.1f}")
            print(f"\tMean countrate per binned pixel: {np.mean(rate):.2e}")
            print(f"\t  Standard deviation: {np.std(rate):.2e}")
            print(f"\tMedian countrate per binned pixel: {np.median(rate):.2e}")
            fig, ax = plt.subplots(figsize=(20,5))
            #vmax = np.mean(binned) * 2
            vmin = np.mean(rate) - 3*np.std(rate)
            if vmin < 0:
                vmin = 0
            vmax = np.mean(rate) + 3*np.std(rate)
            im = ax.imshow(rate, aspect="auto", 
                           origin="lower", cmap="inferno", vmin=vmin, vmax=vmax)
            fig.colorbar(im, label="Counts/s", format="%.2e")
            
            ax.set_title(f"{af['segment']}; HV={af['hv']}; MJD {af['mjdstarts']}-{af['mjdends']}; PHA {spl[0]}-{spl[1]}; X bin={bin_x} Y bin={bin_y}")
            plt.tight_layout()
            pdf.savefig(fig)
        pdf.close()
        print(f"Wrote {pdffile}")

        for k in af.keys():
            if k.startswith("pha"):
                continue
            binned_ims[k] = af[k]
        binned_ims["bin_x"] *= bin_x
        binned_ims["bin_y"] *= bin_y
        binned_ims["bin_pha"] *= bin_pha
        binned_ims["xend"] = b_x1 * bin_x + af["xstart"]
        binned_ims["yend"] = b_y1 * bin_y + af["ystart"]
        binned_ims["phastart"] = phastart
        binned_ims["phaend"] = phaend

    af = asdf.AsdfFile(binned_ims)
    outfile = f"binned_{superdark}"
    af.write_to(outfile)
    print(f"Wrote {outfile}")

    return binned_ims

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(dest="superdark",
                        help="Name of superdark to bin")
    parser.add_argument("-x", "--binx", default=8, type=int,
                        help="Amount to bin in X")
    parser.add_argument("-y", "--biny", default=2, type=int, 
                        help="Amount to bin in X")
    parser.add_argument("-p", "--binpha", default=26, type=int,
                        help="Amount to bin in X")
    parser.add_argument("--phastart", default=3, type=int, 
                        help="Lower limit of PHA range")
    parser.add_argument("--phaend", default=28, type=int, 
                        help="Upper limit of PHA range")
    parser.add_argument("-o", "--outdir", default=".",
                        help="Output directory")
    args = parser.parse_args()
    bin_superdark(args.superdark, args.binx, args.biny, args.binpha,
                  args.phastart, args.phaend, args.outdir)



import yaml
from sqlalchemy.orm import load_only
from sqlalchemy.sql import text
import pandas as pd

from acdc.database.connect_db import load_connection
from acdc.database import darkevents_schema
from acdc.utils.utils import sql_to_df

DBNAME = "dark_events"
COLUMNS = ["id", "xcorr", "ycorr", "pha", "mjd", "hv", "segment", "filename", "proposid"]
TIMING = False

def all_rows(hvtable, returncols=COLUMNS):
    """
    Query the Darks table to get darks for all time for all PHAs.

    Args:
    Returns:
        df (:obj:`pandas.DataFrame`): Query results.
    """
    
    # Connect to database
    session, engine = load_connection(DBNAME)
    tablename = f"DarkEventsHv{hvtable}"
    events_table = getattr(darkevents_schema, tablename)

    # Execute query
    query = session.query(events_table).options(load_only(*returncols))
    results = query.all()

    df = sql_to_df(results, returncols)
    
    return df


def sql_query(hvtable, sqlquery, returncols):
    session, engine = load_connection(DBNAME)
    tablename = f"DarkEventsHv{hvtable}"
    events_table = getattr(darkevents_schema, tablename)

    if returncols == "*":
        returncols = COLUMNS
    elif isinstance(returncols, str):
        returncols = [returncols]

    results = session.execute(sqlquery).fetchall()

# The below query doesn't work if you return a specific column  
#    query = session.query(events_table).from_statement(text(sqlquery))
#    results = query.all()

    d = {}
    for i,col in enumerate(returncols):
        d[col] = [x[i] for x in results]
    df = pd.DataFrame(d)
    
    return df


def equals_query(hvtable, returncols=COLUMNS, **kwargs):
    session, engine = load_connection(DBNAME)
    tablename = f"DarkEventsHv{hvtable}"
    events_table = getattr(darkevents_schema, tablename)

    condition = {col: True for col in COLUMNS}
    if len(kwargs) != 0:
        for col in COLUMNS:
            if col in kwargs:
                colattr = getattr(events_table, col)
                condition[col] = colattr == kwargs[col]

    query = session.query(events_table).options(load_only(*returncols))\
                .filter(condition["xcorr"])\
                .filter(condition["ycorr"])\
                .filter(condition["pha"])\
                .filter(condition["mjd"])\
                .filter(condition["hv"])\
                .filter(condition["segment"])\
                .filter(condition["filename"])\
                .filter(condition["proposid"])
    results = query.all()
    
    df = sql_to_df(results, returncols)
   
    return df 


def range_query(hvtable, returncols=COLUMNS, segment="*", mjdstart=0, mjdend=99999, 
                x0=0, x1=16385, y0=0, y1=1025, pha0=0, pha1=32, hv="*", 
                proposid="*", filename="*"):
    session, engine = load_connection(DBNAME)
    tablename = f"DarkEventsHv{hvtable}"
    events_table = getattr(darkevents_schema, tablename)

    special = ["segment", "hv", "proposid", "filename"]
    condition = {}
    for col in special:
        val = locals()[col] # Get the value of the same-named variable
        if val == "*":
            condition[col] = True
        else:
            colattr = getattr(events_table, col)
            condition[col] = colattr == val

    query = session.query(events_table).options(load_only(*returncols))\
                .filter(condition["segment"])\
                .filter(condition["hv"])\
                .filter(events_table.mjd > mjdstart)\
                .filter(events_table.mjd < mjdend)\
                .filter(events_table.xcorr > x0)\
                .filter(events_table.xcorr < x1)\
                .filter(events_table.ycorr > y0)\
                .filter(events_table.ycorr < y1)\
                .filter(events_table.pha > pha0)\
                .filter(events_table.pha < pha1)\
                .filter(condition["filename"])\
                .filter(condition["proposid"])
    results = query.all()
    
    df = sql_to_df(results, returncols)
   
    return df 




import os
import datetime
import numpy as np
import pandas as pd
from astropy.io import fits

from acdc.database.calculate_dark import get_aperture_region

def sql_to_df(sql_results, returncols):
    # Convert query results (list of attributes) to pandas dataframe
#    d = {col: [getattr(x, col) for x in results] for col in returncols}
    d = {}
    for col in returncols:
        try:
            d[col] = [getattr(x, col) for x in sql_results]
        except:
            pass
    df = pd.DataFrame(d)

    return df

def timefunc(func):  # IN USE
    """
    Decorator to wrap and time functions.

    Parameters
    ----------
    func: func
        Function to be timed

    Returns
    -------
    wrapper: func
        Timed version of the function
    """

    def wrapper(*args, **kw):
        t1 = datetime.datetime.now()
        result = func(*args, **kw)
        t2 = datetime.datetime.now()

        print(f"{func.__name__} executed in {t2 - t1}")

        return result

    return wrapper

def get_binning_pars(af):
    """
    For a given superdark, return the binning information in the spatial     
    directions and PHA.                                       
    """                                                       
    keys = ["bin_pha", "bin_x", "bin_y", "xstart", "xend", "ystart", "yend", 
            "phastart", "phaend"]
    binning = {}                                              
    for k in keys:
        binning[k] = af[k] 
    
    return binning

def bin_coords(xs, ys, bin_x, bin_y, xstart=0, ystart=0, make_int=False, as_float=False):
    """
    Given a list of coordinates in X & Y, transform them into the superdark's
    binned (and possibly offset) coordinate system.
    """
    
    if not isinstance(xs, np.ndarray):
        xs = np.array(xs)
    if not isinstance(ys, np.ndarray):
        ys = np.array(ys)
    if as_float is True:
        xsnew = (xs - xstart) / bin_x
        ysnew = (ys - ystart) / bin_y
    else:
        xsnew = (xs - xstart) // bin_x
        ysnew = (ys - ystart) // bin_y
    if make_int is True:
        xsnew = xsnew.astype(int)
        ysnew = ysnew.astype(int)

    return xsnew, ysnew

def unbin_coords(xs, ys, bin_x, bin_y, xstart=0, ystart=0):
    """
    Given a list of binned coordinates in X & Y, transform them into the the
    unbinned, native coordinate system.
    """
    
    if not isinstance(xs, np.ndarray):
        xs = np.array(xs)
    if not isinstance(ys, np.ndarray):
        ys = np.array(ys)
    xsnew0 = (xs*bin_x) + xstart
    ysnew0 = (ys*bin_y) + ystart
    xsnew1 = xsnew0 + bin_x - 1
    ysnew1 = ysnew0 + bin_y - 1
    return (xsnew0, xsnew1), (ysnew0, ysnew1)

def unbin_image(binned_im, bin_x, bin_y, xstart=0, ystart=0, xend=16384, yend=1024):
    im_perpixel = binned_im / bin_x / bin_y
    unbinned_im = np.zeros(16777216).reshape(1024, 16384)
    xs = np.arange(xstart, xend, bin_x)
    ys = np.arange(ystart, yend, bin_y)
    for i in range(len(xs)-1):
        for j in range(len(ys)-1):
            unbinned_im[ys[j]:ys[j+1], xs[i]:xs[i+1]] = im_perpixel[j,i]
    return unbinned_im

def get_psa_wca(segment, cenwave, lp, binning, pad_psa=[0,0], pad_wca=[0,0]):
    """
    Determine the rows that correspond to the PSA and WCA apertures for a 
    given segment, cenwave, and lifetime position. Return the row indices in 
    the binned superdark coordinate system.

    Arguments:
        segment (str): Get PSA/WCA regions for this segment.
        cenwave (str): Get PSA/WCA regions for this cenwave.
        lp (str): Get PSA/WCA regions for this lifetime position.
        binning (dict): Dictionary that describes the binning information
            in both the spatial and PHA dimensions.
    Returns:
        excluded_rows (array): List of rows to exclude. Corresponds
            to the PSA and WCA regions.
        apertures (dict): The excluded rows for each aperture, PSA and WCA.
    """

    excluded_rows = np.array(())
    apertures = {"PSA": None, "WCA": None}
    pad = {"PSA": pad_psa, "WCA": pad_wca}
    for aperture in apertures:
        aperture_regions = get_aperture_region(cenwave=cenwave, aperture=aperture, segments=[segment], life_adj=[lp])
        box = aperture_regions[segment][f"lp{lp}_{aperture.lower()}_{cenwave}"]
        xmin0, xmax0, ymin0, ymax0 = box                            
        xsnew, ysnew = bin_coords(np.array([xmin0, xmax0]), np.array([ymin0, ymax0]), binning["bin_x"], binning["bin_y"], binning["xstart"], binning["ystart"], as_float=True, make_int=False)
#        ap_xmin, ap_xmax = xsnew # We don't need x coords
        ap_ymin, ap_ymax = ysnew
        ap_ymin = int(np.floor(ap_ymin))
        ap_ymax = int(np.ceil(ap_ymax))
        ap_ymin -= pad[aperture][0]
        ap_ymax += pad[aperture][1]
        rows = np.arange(ap_ymin, ap_ymax)
        apertures[aperture] = rows
        excluded_rows = np.concatenate((excluded_rows, rows))
    excluded_rows = excluded_rows.astype(int)
    return excluded_rows, apertures

def get_background_regions(segment, cenwave, lp=None, date_obs="today", round=True, xtractab=None):
    if xtractab is None:
        if date_obs == "today":
            today = datetime.datetime.now()
            useafter = today.strftime("%Y-%m-%d")

        if "CRDS_PATH" not in os.environ:
            if os.path.exists("/grp/crds/cache"):
                os.environ["CRDS_PATH"] = "/grp/crds/cache"
            else:
                raise AssertionError("CRDS_PATH environment variable must first be defined")
        os.environ["CRDS_SERVER_URL"] = "https://hst-crds.stsci.edu"
        import crds
        current_pmap = crds.get_default_context()
        crds_1dx = crds.getrecommendations(parameters={"INSTRUME": "COS",
                            "DETECTOR": "FUV", "LIFE_ADJ": lp,
                            "OBSTYPE": "SPECTROSCOPIC",
                            "DATE-OBS": useafter, "TIME-OBS": "00:00:00",
                            "CENWAVE": cenwave},
                        reftypes=["xtractab"], context=current_pmap, observatory="hst")
        xtractab = os.path.join(os.environ["CRDS_PATH"], "references/hst/",
                              crds_1dx["xtractab"])
    
    data_1dx = fits.getdata(xtractab, 1)
    ind = np.where((data_1dx["cenwave"] == cenwave) &
                   (data_1dx["segment"] == segment) &
                   (data_1dx["aperture"] == "PSA"))
    xtract_data = data_1dx[ind]
    
    if segment == "FUVA":
        x0 = 1260
        x1 = 15119
    elif segment == "FUVA":
        x0 = 1000
        x1 = 14990
    
    background_d = {}
    for region in [1, 2]:
        b = xtract_data[f"b_bkg{region}"]
        height = xtract_data[f"b_hgt{region}"]
        m = xtract_data["slope"]
        x = np.arange(0, 16384)
        # the y values tracing the line in the middle of the bkgd parallelogram
        bkg_midy = m*x + b 
        bkg_y0 = bkg_midy - height/2.
        bkg_y1 = bkg_midy + height/2.
        if round is True:
            bkg_y0 = np.floor(bkg_y0)
            bkg_y0 = bkg_y0.astype(int)
            bkg_y1 = np.ceil(bkg_y1)
            bkg_y1 = bkg_y1.astype(int)
        background_d[region] = {}
        background_d[region]["bkg_y0"] = bkg_y0
        background_d[region]["bkg_y1"] = bkg_y1

    return background_d
     



import argparse
import datetime
import yaml
import os
import glob
import numpy as np
from timeit import default_timer as timer
from astropy.io import fits
from sqlalchemy import Table
from sqlalchemy.ext.declarative import declarative_base

import acdc.database.create_db as create_db
from acdc.database.connect_db import load_connection
from acdc.database.schema import Solar, Darks
from acdc.database.darkevents_schema import *
from acdc.database.calculate_dark import measure_darkrate, parse_solar_json, get_aperture_region
from acdc.database.saa_models import get_saa_poly
from acdc.database.saa_distance import Distance3dPointTo3dCoords

# For testing purposes only. If TESTING = True, only one value is recorded per
# input dataset to save time. If TIMING = True, recorded runtime for each insert
# is written to STDOUT.
TESTING = False
TIMING  = False

def populate_darkevents(files, dbname="dark_events", db_table="all"):
    """
    Populate the hstcal database. This can be used to add new files
    or ingest all data from scratch.
    
    Args:
        files (array-like): Names of dark exposure corrtags. 
        dbname (str): The name of the database to update. 
        db_table (:obj:`sqlalchemy.Table`): Name of table to update, 
            defaults to 'all', but could be DarkEvents.
    """
    
    # Connect to database.
    session, engine = load_connection(dbname)
    base = declarative_base(engine)
    possible_hvs = [163, 167, 169, 171, 173, 175, 178]
    if db_table == "all":
        print("Inserting events into specific HV tables")
        tables = {}
        for hv in possible_hvs:
            tables[hv] = Table(f"darkeventshv{hv}", base.metadata, autoload=True)
        tables["other"] = Table(f"darkeventshvother", base.metadata, autoload=True)
    else:
        events_table = Table(db_table.__tablename__, base.metadata, autoload=True)

    for fileno,item in enumerate(files):
        loop0 = timer()    
        all_events_rows = []

        with fits.open(item) as hdulist:
            hdr0 = hdulist[0].header
            hdr1 = hdulist[1].header
            data = hdulist[1].data
        segment = hdr0["segment"]
        filename = hdr0["filename"]
        proposid = int(hdr0["proposid"])
        hv = int(hdr1[f"HVLEVEL{segment[-1]}"])
        if db_table == "all":
            if hv not in possible_hvs:
                events_table = tables["other"]
            else:
                events_table = tables[hv]
        mjdstart0 = hdr1["EXPSTART"]
        mjdstart = np.float64(mjdstart0)
        sdqflags = hdr1["sdqflags"]

        good = np.where(data["dq"] & sdqflags == 0)[0]

        timearr0 = data["time"]
        timearr = timearr0.astype(np.float64) 
        for i in good:
            time = mjdstart + timearr[i]/60./60./24.
            event_data = {"xcorr": float(data["xcorr"][i]),
                          "ycorr": float(data["ycorr"][i]),
                          "pha": int(data["pha"][i]),
                          "mjd": float(time),
                          "hv": hv,
                          "segment": segment,
                          "filename": filename,
                          "proposid": proposid}
            all_events_rows.append(event_data)

        insert0 = timer()
        events_table.insert().execute(all_events_rows)
        insert1 = timer()
        print("File {}/{}".format(fileno+1, len(files))) 

        if TIMING is True:
            print("One insert ({} rows) took {:.3g} seconds".format(len(all_events_rows), insert1-insert0))
            print("One file loop took {:.3g} seconds".format(insert1-loop0))
        if TESTING is True:
            print("Updated DarkEvents tables")
            return

    session.commit()
    session.close()
    engine.dispose()

    print("Updated DarkEvents tables")

def populate_solar(files, dbname="cos_dark", db_table=Solar):
    """
    Populate the Solar database table. This can be used to add new files
    or ingest all data from scratch.
    In order to properly get interpolated solar flux values for dark
    observations, Solar table *must* be populated before the Darks table.
    
    Args:
        files (array-like): Names of solar flux files to parse.
        dbname (str): The location of the SQLite database, with full path, e.g.
            /path/to/cos_dark.db 
            If in the current directory, do not include . or ./ 
        db_table (:obj:`sqlalchemy.Table`): Name of table to update, 
            defaults to Solar.
    """
    
    # Connect to database.
    session, engine = load_connection(dbname)
    base = declarative_base(engine)
    solar_table = Table(db_table.__tablename__, base.metadata, autoload=True)


    time,flux = parse_solar_json("observed-solar-cycle-indices.json")

    all_solar_rows = []

    for i in range(len(time)): 
        solar_data = {"time": time[i],
                      "flux": flux[i]}
        all_solar_rows.append(solar_data)

    insert0 = timer()
    solar_table.insert().execute(all_solar_rows)
    insert1 = timer()

    if TIMING is True:
        print("One insert took {:.3g} seconds".format(insert1-insert0))
    if TESTING is True:
        print("Updated table Solar")
        return

    session.commit()
    session.close()
    engine.dispose()
    print("Updated table Solar")

def populate_darks(files, dbname="cos_dark", db_table=Darks):
    """
    Populate the Darks database table. This can be used to add new files
    or ingest all data from scratch.
    In order to properly get interpolated solar flux values for dark
    observations, Solar table *must* be populated before the Darks table.
    
    Args:
        files (array-like): Names of COS dark files to ingest.
        dbname (str): The name of the database to update. 
        db_table (:obj:`sqlalchemy.Table`): Name of table to update,
            defaults to Darks.
    """

    apertures = get_aperture_region()

    # Connect to database.
    session, engine = load_connection(dbname)
    base = declarative_base(engine)
    darks_table = Table(db_table.__tablename__, base.metadata, autoload=True)
    
    # Handle distance to SAA
    saa_dists = {}

    start = datetime.datetime.now()
    for i,item in enumerate(files):
        loop0 = timer()
        all_dark_rows = []

        with fits.open(item) as hdulist:
            hdr0 = hdulist[0].header
            hdr1 = hdulist[1].header
        itemname = os.path.basename(item)
        info, dark = measure_darkrate(item, apertures)

        # The SAA changes over time, so get the right estimation
        saa_lat, saa_lon = get_saa_poly(max(info["time"]))
        
        # Query the Solar table to get solar fluxes near the dark observation
        # dates. Interpolate to get appropriate values. If Solar table is not
        # properly up to date, insert -999 values.
        try:
            solar_q = session.query(Solar.time, Solar.flux)\
                        .filter(Solar.time >= min(info["time"])-2)\
                        .filter(Solar.time <= max(info["time"])+2)
            results = solar_q.all()
            solar_time = [x.time for x in results]
            solar_flux = [x.flux for x in results]
            solar_flux_interp = np.interp(info["time"], solar_time, solar_flux)
        except:
            solar_flux_interp = np.full(len(info["time"]), -999)

        # For each time sampling of the dark observation file, insert a series of
        # values into the Darks table row.
        # Loop over time indices (by default, 25s intervals)
        dark_data0 = {"filename": itemname,
           "segment": info["segment"],
           "exptime": info["exptime"],
           "hv": hdr1["HVLEVEL{}".format(info["segment"][-1])],
           "fileloc": item}
        for j in range(len(info["time"])):
            dark_data1 = dark_data0.copy()
            lat = round(info["latitude"][j], 6)
            lon = round(info["longitude"][j], 6)
            if (lat, lon) not in saa_dists:
                dist = Distance3dPointTo3dCoords(lat, lon, saa_lat, saa_lon)
                saa_dists[(lat, lon)] = dist
            else:
                dist = saa_dists[(lat, lon)]
            dark_data1["latitude"] = lat
            dark_data1["longitude"] = lon
            dark_data1["saa_distance"] = dist
            dark_data1["solar_flux"] = solar_flux_interp[j]
            dark_data1["expstart"] = round(info["time"][j], 6)
            # Loop over regions.
            for region in dark.keys():
                dark_data = dark_data1.copy()
                dark_data["region"] = region
                dark_data["region_area"] = int(dark[region]["region_area"])
                dark_data["xcorr_min"] = int(dark[region]["xcorr_min"])
                dark_data["xcorr_max"] = int(dark[region]["xcorr_max"])
                dark_data["ycorr_min"] = int(dark[region]["ycorr_min"])
                dark_data["ycorr_max"] = int(dark[region]["ycorr_max"])
                dark_data["region_area"] = dark[region]["region_area"]
                for k in range(len(info["pha"])):
                    pha_num = "dark_pha{}".format(info["pha"][k])
                    dark_data[pha_num] = int(dark[region]["darks"][j][k])
                all_dark_rows.append(dark_data)
        insert0 = timer()
        darks_table.insert().execute(all_dark_rows)
        insert1 = timer()
        print("File {}/{}".format(i+1, len(files))) 
        
        if TIMING is True:
            print("One insert took {:.3g} seconds".format(insert1-insert0))
            print("One file loop took {:.3g} seconds".format(insert1-loop0))
        if TESTING is True:
            print("Updated table Darks")
            return

    session.commit()
    session.close()
    engine.dispose()
    end = datetime.datetime.now()
    print("Updated table Darks")

def update_cos_dark(dbname):
    with open("settings.yaml", "r") as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)
        dbsettings = settings["dbsettings"][dbname]
    if not os.path.exists(dbsettings["loc"]):
        create_db.create_db(dbname)
    start = datetime.datetime.now()
    print("Start time: {}".format(start))
    solar_files = glob.glob(os.path.join(settings["solar_dir"]))
    populate_solar(solar_files)
    files = glob.glob(os.path.join(settings["dark_dir"]))
    populate_darks(files)
    end = datetime.datetime.now()
    print("Start time: {}".format(start))
    print("End time: {}".format(end))
    print("Total time: {}".format(end-start))


def update_dark_events(tablename=None):
    with open("settings.yaml", "r") as f:
        settings = yaml.load(f, Loader=yaml.SafeLoader)
    start = datetime.datetime.now()
    print("Start time: {}".format(start))
    files = glob.glob(os.path.join(settings["dark_dir"]))
    if tablename == None:
        tablename = "all"
    populate_darkevents(files, db_table=tablename)
    end = datetime.datetime.now()
    print("Start time: {}".format(start))
    print("End time: {}".format(end))
    print("Total time: {}".format(end-start))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dbname", help="Database to update")
    parser.add_argument("-t", "--tablename", default=None,
                        help="Table(s) to update")
    args = parser.parse_args()
    dbname = args.dbname
    if dbname == "cos_dark":
        update_cos_dark(dbname)
    elif dbname == "dark_events":
        update_dark_events(args.tablename)
    else:
        print("Invalid database name supplied: currently supported databases are cos_dark and dark_events")



import argparse
import yaml

from acdc.database.connect_db import load_connection

def create_db(dbname="cos_dark"):
    """
    Create the database, with name specified from `settings.yaml`.

    Args:
        dbname (str): Name of database to create with empty tables.
            Currently cos_dark and dark_events are supported. 
    """

    if dbname == "cos_dark":
        from schema import Base, Darks, Solar
    elif dbname == "dark_events":
        from darkevents_schema import Base, DarkEvents
    session, engine = load_connection(dbname)
    # It's important that the Base from appropriate be schema be used, imported above
    Base.metadata.create_all(engine)
    print(f"Created database & tables for {dbname}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--db", default="cos_dark",
                        help="Name of database to create")
    args = parser.parse_args()
    create_db(args.db)



from sqlalchemy import Index

from acdc.database.connect_db import load_connection
from acdc.database.darkevents_schema import DarkEvents

# Connect to database
session, engine = load_connection("dark_events")

hv_index = Index("hv_index", DarkEvents.hv)

hv_index.create(bind=engine)



import argparse
from acdc.dark_correction import Acdc


def run_acdc(indir, darkcorr_outdir, x1d_outdir=None,
             superdark_dir=None, binned=True, hv=None, 
             segment=None, overwrite=False, exclude_airglow=False,
             calibrate=True):
    """Wrapper script to perform custom dakr correction on an input directory.
    
    Args:
        indir (str): Input directory that houses corrtags to correct.
        darkcorr_outdir (str): Custom dark-corrected corrtags will be written here.
        x1d_outdir (str): Custom dark-corrected x1ds will be written here.
        binned (Bool): (Optional) If True, pre-binned superdarks will be used.
        hv (str): (Optional) Process corrtags of this HV only.
        segment (str): (Optional) Process corrtags of this segment only.
    """
    A = Acdc(indir=indir, darkcorr_outdir=darkcorr_outdir, x1d_outdir=x1d_outdir,
             superdark_dir=superdark_dir,
             binned=binned, segment=segment, hv=hv, overwrite=overwrite,
             exclude_airglow=exclude_airglow, calibrate=calibrate)
    A.custom_dark_correction()
    A.calibrate_corrtags()
    print(f"\nFINISHED: Products in {A.x1d_outdir}")

def acdc_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--indir",
                        help="Path to science corrtags")
    parser.add_argument("-o", "--darkcorr_outdir",
                        help="Name of directory to write model superdarks and custom corrtags to")
    parser.add_argument("--x1d_outdir", default=None,
                        help="Name of directory to write 1D spectra to")
    parser.add_argument("-d", "--darkdir", default=None,
                        help="Directory that houses superdarks to use")
    parser.add_argument("--unbinned", dest="binned", default=True,
                        action="store_false",
                        help="Toggle to indicate that supplied superdarks are binned")
    parser.add_argument("-c", "--clobber", default=False,
                        action="store_true",
                        help="Toggle to overwrite any existing products")
    parser.add_argument("--hv", default=None,
                        help="HV to filter corrtags by")
    parser.add_argument("--segment", default=None,
                        help="Segment to filter corrtags by")
    parser.add_argument("--airglow", default=False, dest="exclude_airglow",
                        action="store_true",
                        help="Toggle to exclude LyA and OI airglow from science data scaling") 
    parser.add_argument("--nocal", default=True, dest="calibrate",
                        action="store_false",
                        help="Toggle to turn off CalCOS calibration") 
    args = parser.parse_args()
    run_acdc(args.indir, args.darkcorr_outdir, args.x1d_outdir, 
             args.darkdir, args.binned, args.hv, 
             args.segment, args.clobber, args.exclude_airglow, args.calibrate)


if __name__ == "__main__":
	acdc_parser()



import numpy as np
import math

# Model 31 from https://github.com/spacetelescope/costools/blob/master/costools/saamodel.py
COS_FUV_MODEL = [
	(-28.3,         14.0),
	(-27.5,         15.0),
	(-26.1,         13.0),
	(-19.8,          1.5),
	( -9.6,        341.0),
	( -7.6,        330.4),
	( -6.0,        318.8),
	( -7.9,        297.2),
	(-12.0,        286.1),
	(-17.1,        279.9),
	(-20.3,        277.5),
	(-23.5,        276.5),
	(-26.0,        276.4),
	(-28.6,        276.7)]

DEGtoRAD = math.pi / 180.
TWOPI = 2. * math.pi
# This cutoff is based on current models in saamodel.py; this is used when
# finding the middle of the SAA region (middle_SAA).
SAA_LONGITUDE_CUTOFF = 200.



def testWithinSAA(hst, vertices, middle_SAA):
    """Test whether HST is within the polygon for an SAA contour.
    
    Args:
        hst (array_like): Unit vector pointing from the center of the Earth toward the
            location of HST at a particular time.
        vertices (array_like, shape (nvertices,3)): vertices[i] is a unit vector 
            from the center of the Earth toward
            vertex number i of a polygon that defines one of the SAA contour.
        middle_SAA (array_like): Unit vector from the center of the Earth toward a point near the
            middle of the SAA region.  This is for making a quick check that
            hst is close enough to the SAA contour to be worth making a
            detailed check.
    
    Returns:
        bool: True if hst is within the SAA contour defined by vertices.
    """

    # This test is primarily to exclude points that are diametrically
    # opposite to the SAA contour (because this would not be caught by
    # the code below!), but it should also save unnecessary arithmetic
    # most of the time.
    if np.dot(hst, middle_SAA) < 0.:
        return False

    nvertices = len(vertices)

    sin_lat_hst = hst[2]
    cos_lat_hst = math.sqrt(1. - sin_lat_hst**2)
    cos_long_hst = hst[0] / cos_lat_hst
    sin_long_hst = hst[1] / cos_lat_hst

    # vertices rotated to put hst in the x-z plane
    v_rot = vertices.copy()
    v_rot[:,0] =  vertices[:,0] * cos_long_hst + vertices[:,1] * sin_long_hst
    v_rot[:,1] = -vertices[:,0] * sin_long_hst + vertices[:,1] * cos_long_hst

    # v_rot rotated to put hst on the x axis
    v_rotrot = v_rot.copy()
    v_rotrot[:,0] =  v_rot[:,0] * cos_lat_hst + v_rot[:,2] * sin_lat_hst
    v_rotrot[:,2] = -v_rot[:,0] * sin_lat_hst + v_rot[:,2] * cos_lat_hst

    azimuth = np.arctan2(v_rotrot[:,2], v_rotrot[:,1])
    azimuth = np.where(azimuth < 0., azimuth + TWOPI, azimuth)

    delta_az = azimuth[1:] - azimuth[0:-1]
    delta_az = np.where(delta_az < -np.pi, delta_az + TWOPI, delta_az)
    delta_az = np.where(delta_az > np.pi, delta_az - TWOPI, delta_az)

    sum_delta_az = delta_az.sum()

    return not (sum_delta_az < 0.1 and sum_delta_az > -0.1)


def saaFilter(longitude_col, latitude_col, model=COS_FUV_MODEL):
    """Flag within the specified SAA contour as bad.
    
    Args:
        model (int): The SAA model number.  Currently these range from 2 to 32
            inclusive.  (Models 0 and 1 are radio frequence interference
            contours.)
    
    Returns:
        flag (array_like): This is a boolean array, one element for each row of the
            TIMELINE table.  True means that HST was within the SAA
            contour (specified by model) at the time corresponding to
            the TIMELINE row.
    """

    nelem = len(longitude_col)

    flag = np.zeros(nelem, dtype=np.bool8)

    model_vertices = model
    model_vertices.append(model_vertices[0])        # make a closed loop
    nvertices = len(model_vertices)
    # will be unit vectors from center of Earth pointing toward vertices
    vertices = np.zeros((nvertices, 3), dtype=np.float64)
    minmax_long = [720., -360.]     # will be minimum, maximum longitudes
    minmax_lat = [90., -90.]        # will be minimum, maximum latitudes
    for i in range(nvertices):
        (latitude, longitude) = model_vertices[i]
        vertices[i] = toRect(longitude, latitude)   # change the order
        if longitude < SAA_LONGITUDE_CUTOFF:
            longitude += 360.
        minmax_long[0] = min(longitude, minmax_long[0])
        minmax_long[1] = max(longitude, minmax_long[1])
        minmax_lat[0] = min(latitude, minmax_lat[0])
        minmax_lat[1] = max(latitude, minmax_lat[1])
    middle_long = (minmax_long[0] + minmax_long[1]) / 2.
    middle_lat = (minmax_lat[0] + minmax_lat[1]) / 2.
    middle_SAA = toRect(middle_long, middle_lat)

    # for each row in TIMELINE table
    for k in range(nelem):
        hst = toRect(longitude_col[k], latitude_col[k])
        flag[k] = testWithinSAA(hst, vertices, middle_SAA)

    return flag


def toRect(longitude, latitude):
    """Convert longitude and latitude to rectangular coordinates.
    
    Args:
        longitude (float): longitude in degrees.
        latitude (float): latitude in degrees.
    
    Returns:
        rect (array-like): Unit vector in rectangular coordinates.
    """

    rect = np.array([1.0, 0.0, 0.0], dtype=np.float64)
    longitude *= DEGtoRAD
    latitude *= DEGtoRAD

    rect[0] = math.cos(latitude) * math.cos(longitude)
    rect[1] = math.cos(latitude) * math.sin(longitude)
    rect[2] = math.sin(latitude)

    return rect

